
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_limit_active_intervals_constraint.c                    */
/*  DESCRIPTION:  A limit active intervals constraint                        */
/*                                                                           */
/*****************************************************************************/
#include "khe_interns.h"

#define DEBUG1 0


/*****************************************************************************/
/*                                                                           */
/*  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT - limit active intervals constraint*/
/*                                                                           */
/*****************************************************************************/

struct khe_limit_active_intervals_constraint_rec {
  INHERIT_CONSTRAINT
  ARRAY_KHE_RESOURCE_GROUP	resource_groups;	/* applies to        */
  ARRAY_KHE_RESOURCE		resources;		/* applies to        */
  int				history_before;		/* history: ai       */
  int				history_after;		/* history: ci       */
  HA_ARRAY_INT			history;		/* history: xi       */
  /* int			history_dft_before; */	/* default history   */
  /* int			history_dft_after; */	/* default history   */
  /* ARRAY_INT			history_before; */	/* histories         */
  /* ARRAY_INT			history_after; */	/* histories         */
  KHE_TIME_GROUP		applies_to_tg;		/* applies to tg     */
  HA_ARRAY_INT			applies_to_offsets;	/* applies_to offsets*/
  int				minimum;		/* minimum           */
  int				maximum;		/* maximum           */
  ARRAY_KHE_TIME_GROUP		time_groups;		/* time groups       */
  ARRAY_KHE_POLARITY		polarities;		/* polarities        */
  bool				all_positive;		/* if all positive   */
  bool				all_negative;		/* if all negative   */
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "construction and query"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitActiveIntervalsConstraintMake(KHE_INSTANCE ins, char *id,   */
/*    char *name, bool required, int weight, KHE_COST_FUNCTION cf,           */
/*    KHE_TIME_GROUP applies_to_tg, int minimum, int maximum,                */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT *c)                              */
/*                                                                           */
/*  Make a new limit active intervals constraint with these attributes, add  */
/*  it to ins, and return it.                                                */
/*                                                                           */
/*****************************************************************************/

bool KheLimitActiveIntervalsConstraintMake(KHE_INSTANCE ins, char *id,
  char *name, bool required, int weight, KHE_COST_FUNCTION cf,
  KHE_TIME_GROUP applies_to_tg, int minimum, int maximum,
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT *c)
{
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT res;  KHE_CONSTRAINT cc;  HA_ARENA a;
  HnAssert(!KheInstanceComplete(ins),
    "KheLimitActiveIntervalsConstraintMake called after KheInstanceMakeEnd");
  HnAssert(KheInstanceModel(ins) == KHE_MODEL_EMPLOYEE_SCHEDULE,
    "KheLimitActiveIntervalsConstraintMake called in high school model");
  if( id != NULL && KheInstanceRetrieveConstraint(ins, id, &cc) )
  {
    *c = NULL;
    return false;
  }
  a = KheInstanceArena(ins);
  HaMake(res, a);
  KheConstraintInitCommonFields((KHE_CONSTRAINT) res,
    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT_TAG, ins, id, name, required,
    weight, cf, a);
  HaArrayInit(res->resource_groups, a);
  HaArrayInit(res->resources, a);
  res->history_before = -1;  /* signals undefined */
  res->history_after = -1;  /* signals undefined */
  HaArrayInit(res->history, a);
  /* res->history_dft_before = -1; */  /* signals undefined */
  /* res->history_dft_after = -1; */  /* signals undefined */
  /* HaArrayInit(res->history_before, a); */
  /* HaArrayInit(res->history_after, a); */
  res->applies_to_tg = applies_to_tg;
  HaArrayInit(res->applies_to_offsets, a);
  HaArrayAddLast(res->applies_to_offsets, -1);  /* signals not initialized */
  res->minimum = minimum;
  res->maximum = maximum;
  HaArrayInit(res->time_groups, a);
  HaArrayInit(res->polarities, a);
  res->all_positive = true;
  res->all_negative = true;
  KheInstanceAddConstraint(ins, (KHE_CONSTRAINT) res);
  *c = res;
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_TIME_GROUP KheLimitActiveIntervalsConstraintAppliesToTimeGroup(      */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)                               */
/*                                                                           */
/*  Return the applies-to time group attribute of c (it may be NULL).        */
/*                                                                           */
/*****************************************************************************/

KHE_TIME_GROUP KheLimitActiveIntervalsConstraintAppliesToTimeGroup(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)
{
  return c->applies_to_tg;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheLimitActiveIntervalsConstraintMinimum(                            */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)                               */
/*                                                                           */
/*  Return the minimum attribute of c.                                       */
/*                                                                           */
/*****************************************************************************/

int KheLimitActiveIntervalsConstraintMinimum(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)
{
  return c->minimum;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheLimitActiveIntervalsConstraintMaximum(                            */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)                               */
/*                                                                           */
/*  Return the maximum attribute of c.                                       */
/*                                                                           */
/*****************************************************************************/

int KheLimitActiveIntervalsConstraintMaximum(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)
{
  return c->maximum;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheLimitActiveIntervalsConstraintAppliesToTimes(                     */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)                               */
/*                                                                           */
/*  Return the number of applies-to offsets of c.                            */
/*                                                                           */
/*****************************************************************************/

int KheLimitActiveIntervalsConstraintAppliesToOffsetCount(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)
{
  if( HaArrayCount(c->applies_to_offsets) == 1 &&
      HaArrayFirst(c->applies_to_offsets) == -1 )
    HnAbort("KheLimitActiveIntervalsConstraintAppliesToOffsetCount"
      " called before KheInstanceMakeEnd");
  return HaArrayCount(c->applies_to_offsets);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheLimitActiveIntervalsConstraintAppliesToOffset(                    */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int i)                        */
/*                                                                           */
/*  Return the i'th applies-to offset of c.                                  */
/*                                                                           */
/*****************************************************************************/

int KheLimitActiveIntervalsConstraintAppliesToOffset(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int i)
{
  return HaArray(c->applies_to_offsets, i);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheLimitActiveIntervalsResourceCount(                                */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)                               */
/*                                                                           */
/*  Return the number of resources that c applies to, including resources    */
/*  in resource groups.                                                      */
/*                                                                           */
/*****************************************************************************/

static int KheLimitActiveIntervalsResourceCount(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)
{
  int i, res;  KHE_RESOURCE_GROUP rg;
  res = HaArrayCount(c->resources);
  HaArrayForEach(c->resource_groups, rg, i)
    res += KheResourceGroupResourceCount(rg);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheLimitActiveIntervalsConstraintAppliesToCount(                     */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)                               */
/*                                                                           */
/*  Return the number of points of application of c.                         */
/*                                                                           */
/*****************************************************************************/

int KheLimitActiveIntervalsConstraintAppliesToCount(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)
{
  return KheLimitActiveIntervalsResourceCount(c) *
    KheLimitActiveIntervalsConstraintAppliesToOffsetCount(c);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheLimitActiveIntervalsConstraintFinalize(                          */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)                               */
/*                                                                           */
/*  Finalize c, since KheInstanceMakeEnd has been called.                    */
/*                                                                           */
/*****************************************************************************/

void KheLimitActiveIntervalsConstraintFinalize(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)
{
  int max_index, base_index, tg_count, i, time_count, index;  KHE_TIME_GROUP tg;

  /* finalize c->applies_to_offsets */
  HaArrayClear(c->applies_to_offsets);
  if( c->applies_to_tg == NULL || KheTimeGroupTimeCount(c->applies_to_tg) == 0 )
    HaArrayAddLast(c->applies_to_offsets, 0);
  else if( KheTimeGroupTimeCount(c->applies_to_tg) >= 1 )
  {
    /* find max_index, the index of the largest time in c->time_groups, or -1 */
    max_index = -1;
    HaArrayForEach(c->time_groups, tg, i)
    {
      tg_count = KheTimeGroupTimeCount(tg);
      if( tg_count > 0 )
      {
	index = KheTimeIndex(KheTimeGroupTime(tg, tg_count - 1));
	if( index > max_index )
	  max_index = index;
      }
    }

    /* add legal offsets to applies_to_offsets */
    time_count = KheInstanceTimeCount(c->instance);
    base_index = KheTimeIndex(KheTimeGroupTime(c->applies_to_tg, 0));
    for( i = 0;  i < KheTimeGroupTimeCount(c->applies_to_tg);  i++ )
    {
      index = KheTimeIndex(KheTimeGroupTime(c->applies_to_tg, i));
      if( max_index + (index - base_index) >= time_count )
	break;
      HaArrayAddLast(c->applies_to_offsets, index - base_index);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  int KheLimitActiveIntervalsConstraintDensityCount(                       */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)                               */
/*                                                                           */
/*  Return the density count of c; just the applies to count in this case.   */
/*                                                                           */
/*****************************************************************************/

int KheLimitActiveIntervalsConstraintDensityCount(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)
{
  if( DEBUG1 )
    fprintf(stderr,
      "KheLimitActiveIntervalsConstraintDensityCount returning %d",
      KheLimitActiveIntervalsConstraintAppliesToCount(c));
  return KheLimitActiveIntervalsConstraintAppliesToCount(c);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "resource groups"                                              */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheLimitActiveIntervalsConstraintAddResourceGroup(                  */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE_GROUP rg)        */
/*                                                                           */
/*  Add rg to c.                                                             */
/*                                                                           */
/*****************************************************************************/

void KheLimitActiveIntervalsConstraintAddResourceGroup(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE_GROUP rg)
{
  int i;  KHE_RESOURCE r;
  HaArrayAddLast(c->resource_groups, rg);
  for( i = 0;  i < KheResourceGroupResourceCount(rg);  i++ )
  {
    r = KheResourceGroupResource(rg, i);
    KheResourceAddConstraint(r, (KHE_CONSTRAINT) c);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  int KheLimitActiveIntervalsConstraintResourceGroupCount(                 */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)                               */
/*                                                                           */
/*  Return the number of resource groups in c.                               */
/*                                                                           */
/*****************************************************************************/

int KheLimitActiveIntervalsConstraintResourceGroupCount(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)
{
  return HaArrayCount(c->resource_groups);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_RESOURCE_GROUP KheLimitActiveIntervalsConstraintResourceGroup(       */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int i)                        */
/*                                                                           */
/*  Return the i'th resource group of c.                                     */
/*                                                                           */
/*****************************************************************************/

KHE_RESOURCE_GROUP KheLimitActiveIntervalsConstraintResourceGroup(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int i)
{
  return HaArray(c->resource_groups, i);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "resources"                                                    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheLimitActiveIntervalsConstraintAddResource(                       */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r)               */
/*                                                                           */
/*  Add r to c.                                                              */
/*                                                                           */
/*****************************************************************************/

void KheLimitActiveIntervalsConstraintAddResource(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r)
{
  HaArrayAddLast(c->resources, r);
  KheResourceAddConstraint(r, (KHE_CONSTRAINT) c);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheLimitActiveIntervalsConstraintResourceCount(                      */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)                               */
/*                                                                           */
/*  Return the number of resources of c.                                     */
/*                                                                           */
/*****************************************************************************/

int KheLimitActiveIntervalsConstraintResourceCount(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)
{
  return HaArrayCount(c->resources);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_RESOURCE KheLimitActiveIntervalsConstraintResource(                  */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int i)                        */
/*                                                                           */
/*  Return the i'th resource of c.                                           */
/*                                                                           */
/*****************************************************************************/

KHE_RESOURCE KheLimitActiveIntervalsConstraintResource(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int i)
{
  return HaArray(c->resources, i);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "resource history"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheLimitActiveIntervalsConstraintAddHistoryBefore(                  */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int val)                      */
/*                                                                           */
/*  Add a history before value to c.                                         */
/*                                                                           */
/*****************************************************************************/

void KheLimitActiveIntervalsConstraintAddHistoryBefore(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int val)
{
  HnAssert(KheInstanceModel(c->instance) == KHE_MODEL_EMPLOYEE_SCHEDULE,
    "KheLimitActiveIntervalsConstraintAddHistoryBefore: wrong model");
  HnAssert(c->history_before == -1,
    "KheLimitActiveIntervalsConstraintAddHistoryBefore called twice");
  HnAssert(val>=0, "KheLimitActiveIntervalsConstraintAddHistoryBefore: val < 0");
  c->history_before = val;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitActiveIntervalsConstraintHistoryBefore(                     */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int *val)                     */
/*                                                                           */
/*  Retrieve a history before value from c.                                  */
/*                                                                           */
/*****************************************************************************/

bool KheLimitActiveIntervalsConstraintHistoryBefore(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int *val)
{
  *val = c->history_before;
  return c->history_before != -1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheLimitActiveIntervalsConstraintAddHistoryAfter(                   */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int val)                      */
/*                                                                           */
/*  Add a history after value to c.                                          */
/*                                                                           */
/*****************************************************************************/

void KheLimitActiveIntervalsConstraintAddHistoryAfter(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int val)
{
  HnAssert(KheInstanceModel(c->instance) == KHE_MODEL_EMPLOYEE_SCHEDULE,
    "KheLimitActiveIntervalsConstraintAddHistoryAfter: wrong model");
  HnAssert(c->history_after == -1,
    "KheLimitActiveIntervalsConstraintAddHistoryAfter called twice");
  HnAssert(val>=0, "KheLimitActiveIntervalsConstraintAddHistoryAfter: val < 0");
  c->history_after = val;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitActiveIntervalsConstraintHistoryAfter(                      */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int *val)                     */
/*                                                                           */
/*  Retrieve a history after value from c.                                   */
/*                                                                           */
/*****************************************************************************/

bool KheLimitActiveIntervalsConstraintHistoryAfter(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int *val)
{
  *val = c->history_after;
  return c->history_after != -1;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheLimitActiveIntervalsConstraintAddHistory(                        */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r, int val)      */
/*                                                                           */
/*  Add a history value to c.                                                */
/*                                                                           */
/*****************************************************************************/

void KheLimitActiveIntervalsConstraintAddHistory(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r, int val)
{
  int index;
  HnAssert(KheInstanceModel(c->instance) == KHE_MODEL_EMPLOYEE_SCHEDULE,
    "KheLimitActiveIntervalsConstraintAddHistory: wrong model");
  HnAssert(val >= 0, "KheLimitActiveIntervalsConstraintAddHistory: val < 0");
  index = KheResourceInstanceIndex(r);
  HaArrayFill(c->history, index + 1, -1);
  HnAssert(HaArray(c->history, index) == -1,
    "KheLimitActiveIntervalsConstraintAddHistory: history added twice");
  HaArrayPut(c->history, index, val);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitActiveIntervalsConstraintHistory(                           */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r, int *val)     */
/*                                                                           */
/*  Retrieve a history value from c.                                         */
/*                                                                           */
/*****************************************************************************/

bool KheLimitActiveIntervalsConstraintHistory(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r, int *val)
{
  int index, res;
  index = KheResourceInstanceIndex(r);
  if( HaArrayCount(c->history) <= index )
  {
    *val = -1;
    return false;
  }
  res = HaArray(c->history, index);
  if( res == -1 )
  {
    *val = -1;
    return false;
  }
  else
  {
    *val = res;
    return true;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheLimitActiveIntervalsConstraintAddHistoryDefaultBefore(           */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int val)                      */
/*                                                                           */
/*  Add a default before value for history.                                  */
/*                                                                           */
/*****************************************************************************/

/* ***
void KheLimitActiveIntervalsConstraintAddHistoryDefaultBefore(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int val)
{
  HnAssert(c->history_dft_before == -1,
    "KheLimitActiveIntervalsConstraintAddHistoryDefaultBefore called twice");
  HnAssert(val >= 0,
    "KheLimitActiveIntervalsConstraintAddHistoryDefaultBefore: val < 0");
  c->history_dft_before = val;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheLimitActiveIntervalsConstraintAddHistoryDefaultAfter(            */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int val)                      */
/*                                                                           */
/*  Add a default after value for history.                                   */
/*                                                                           */
/*****************************************************************************/

/* ***
void KheLimitActiveIntervalsConstraintAddHistoryDefaultAfter(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int val)
{
  HnAssert(c->history_dft_after == -1,
    "KheLimitActiveIntervalsConstraintAddHistoryDefaultAfter called twice");
  HnAssert(val >= 0,
    "KheLimitActiveIntervalsConstraintAddHistoryDefaultAfter: val < 0");
  c->history_dft_after = val;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitActiveIntervalsConstraintRetrieveHistoryDefaultBefore(      */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int *val)                     */
/*                                                                           */
/*  Retrieve a default before value for history.                             */
/*                                                                           */
/*****************************************************************************/

/* ***
bool KheLimitActiveIntervalsConstraintRetrieveHistoryDefaultBefore(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int *val)
{
  *val = c->history_dft_before;
  return c->history_dft_before >= 0;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitActiveIntervalsConstraintRetrieveHistoryDefaultAfter(       */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int *val)                     */
/*                                                                           */
/*  Retrieve a default after value for history.                              */
/*                                                                           */
/*****************************************************************************/

/* ***
bool KheLimitActiveIntervalsConstraintRetrieveHistoryDefaultAfter(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int *val)
{
  *val = c->history_dft_after;
  return c->history_dft_after >= 0;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheLimitActiveIntervalsConstraintAddHistoryBefore(                  */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r, int val)      */
/*                                                                           */
/*  Add a before value for history.                                          */
/*                                                                           */
/*****************************************************************************/

/* ***
void KheLimitActiveIntervalsConstraintAddHistoryBefore(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r, int val)
{
  int index;
  index = KheResourceInstanceIndex(r);
  HnAssert(val >= 0,
    "KheLimitActiveIntervalsConstraintAddHistoryBefore: val < 0");
  HaArrayFill(c->history_before, index + 1, -1);
  HnAssert(HaArray(c->history_before, index) == -1,
    "KheLimitActiveIntervalsConstraintAddHistoryBefore: history added twice");
  HaArrayPut(c->history_before, index, val);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheLimitActiveIntervalsConstraintAddHistoryAfter(                   */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r, int val)      */
/*                                                                           */
/*  Add an after value for history.                                          */
/*                                                                           */
/*****************************************************************************/

/* ***
void KheLimitActiveIntervalsConstraintAddHistoryAfter(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r, int val)
{
  int index;
  index = KheResourceInstanceIndex(r);
  HnAssert(val >= 0,
    "KheLimitActiveIntervalsConstraintAddHistoryAfter: val < 0");
  HaArrayFill(c->history_after, index + 1, -1);
  HnAssert(HaArray(c->history_after, index) == -1,
    "KheLimitActiveIntervalsConstraintAddHistoryAfter: history added twice");
  HaArrayPut(c->history_after, index, val);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitActiveIntervalsConstraintRetrieveHistoryBefore(             */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r, int *val)     */
/*                                                                           */
/*  Retrieve a before value for history.                                     */
/*                                                                           */
/*****************************************************************************/

/* ***
bool KheLimitActiveIntervalsConstraintRetrieveHistoryBefore(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r, int *val)
{
  int index, res;
  index = KheResourceInstanceIndex(r);
  if( HaArrayCount(c->history_before) <= index )
  {
    *val = 0;
    return false;
  }
  res = HaArray(c->history_before, index);
  if( res >= 0 )
  {
    *val = res;
    return true;
  }
  else
  {
    *val = 0;
    return false;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitActiveIntervalsConstraintRetrieveHistoryAfter(              */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r, int *val)     */
/*                                                                           */
/*  Retrieve an after value for history.                                     */
/*                                                                           */
/*****************************************************************************/

/* ***
bool KheLimitActiveIntervalsConstraintRetrieveHistoryAfter(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r, int *val)
{
  int index, res;
  index = KheResourceInstanceIndex(r);
  if( HaArrayCount(c->history_after) <= index )
  {
    *val = 0;
    return false;
  }
  res = HaArray(c->history_after, index);
  if( res >= 0 )
  {
    *val = res;
    return true;
  }
  else
  {
    *val = 0;
    return false;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheLimitActiveIntervalsConstraintHistoryBefore(                      */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r)               */
/*                                                                           */
/*  Return r's before value for history:  either a value given specifically  */
/*  for r, or else the constraint default value, or else 0.                  */
/*                                                                           */
/*****************************************************************************/

/* ***
int KheLimitActiveIntervalsConstraintHistoryBefore(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r)
{
  int v;
  if( KheLimitActiveIntervalsConstraintRetrieveHistoryBefore(c, r, &v) )
    return v;
  else if( KheLimitActiveIntervalsConstraintRetrieveHistoryDefaultBefore(c,&v) )
    return v;
  else
    return 0;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheLimitActiveIntervalsConstraintHistoryAfter(                       */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r)               */
/*                                                                           */
/*  Return r's after value for history:  either a value given specifically   */
/*  for r, or else the constraint default value, or else 0.                  */
/*                                                                           */
/*****************************************************************************/

/* ***
int KheLimitActiveIntervalsConstraintHistoryAfter(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r)
{
  int v;
  if( KheLimitActiveIntervalsConstraintRetrieveHistoryAfter(c, r, &v) )
    return v;
  else if( KheLimitActiveIntervalsConstraintRetrieveHistoryDefaultAfter(c, &v) )
    return v;
  else
    return 0;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheLimitActiveIntervalsConstraintAddHistory(                        */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r, int val)      */
/*                                                                           */
/*  Add history (r, val) to c.                                               */
/*                                                                           */
/*****************************************************************************/

/* ***
void KheLimitActiveIntervalsConstraintAddHistory(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r, int val)
{
  int index;
  index = KheResourceInstanceIndex(r);
  HnAssert(val >= 0, "KheLimitActiveIntervalsConstraintAddHistory: val < 0");
  HaArrayFill(c->resource_histories, index + 1, -1);
  HnAssert(HaArray(c->resource_histories, index) == -1,
    "KheLimitActiveIntervalsConstraintAddHistory: history added twice");
  HaArrayPut(c->resource_histories, index, val);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitActiveIntervalsConstraintRetrieveHistory(                   */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r, int *val)     */
/*                                                                           */
/*  Retrieve a history value from c.                                         */
/*                                                                           */
/*****************************************************************************/

/* ***
bool KheLimitActiveIntervalsConstraintRetrieveHistory(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r, int *val)
{
  int index, res;
  index = KheResourceInstanceIndex(r);
  HaArrayFill(c->resource_histories, index + 1, -1);
  res = HaArray(c->resource_histories, index);
  if( res >= 0 )
  {
    *val = res;
    return true;
  }
  else
  {
    *val = 0;
    return false;
  }
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "time groups"                                                  */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheLimitActiveIntervalsConstraintAddTimeGroup(                      */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_TIME_GROUP tg,            */
/*    KHE_POLARITY po)                                                       */
/*                                                                           */
/*  Add tg with polarity po to c.                                            */
/*                                                                           */
/*****************************************************************************/

void KheLimitActiveIntervalsConstraintAddTimeGroup(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_TIME_GROUP tg, KHE_POLARITY po)
{
  HnAssert(tg != NULL,
    "KheLimitActiveIntervalsConstraintAddTimeGroup: tg is NULL");
  if( KheInstanceModel(c->instance) == KHE_MODEL_HIGH_SCHOOL_TIMETABLE )
    HnAssert(po == KHE_POSITIVE, "KheLimitActiveIntervalsConstraintAddTimeGroup:"
     " KHE_NEGATIVE in high school model");
  HaArrayAddLast(c->time_groups, tg);
  HaArrayAddLast(c->polarities, po);
  if( po == KHE_POSITIVE )
    c->all_negative = false;
  else
    c->all_positive = false;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheLimitActiveIntervalsConstraintTimeGroupCount(                     */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)                               */
/*                                                                           */
/*  Return the number of time groups of c.                                   */
/*                                                                           */
/*****************************************************************************/

int KheLimitActiveIntervalsConstraintTimeGroupCount(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)
{
  return HaArrayCount(c->time_groups);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_TIME_GROUP KheLimitActiveIntervalsConstraintTimeGroup(               */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int i, int offset,            */
/*    KHE_POLARITY *po)                                                      */
/*                                                                           */
/*  Return the i'th time group of c.                                         */
/*                                                                           */
/*****************************************************************************/

KHE_TIME_GROUP KheLimitActiveIntervalsConstraintTimeGroup(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, int i, int offset, KHE_POLARITY *po)
{
  KHE_TIME_GROUP tg;
  *po = HaArray(c->polarities, i);
  tg = HaArray(c->time_groups, i);
  return offset == 0 ? tg : KheTimeGroupNeighbour(tg, offset);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitActiveIntervalsConstraintAllPositive(                       */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)                               */
/*                                                                           */
/*  Return true if all time groups are positive.                             */
/*                                                                           */
/*****************************************************************************/

bool KheLimitActiveIntervalsConstraintAllPositive(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)
{
  return c->all_positive;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitActiveIntervalsConstraintAllNegative(                       */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)                               */
/*                                                                           */
/*  Return true if all time groups are negative.                             */
/*                                                                           */
/*****************************************************************************/

bool KheLimitActiveIntervalsConstraintAllNegative(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c)
{
  return c->all_negative;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "monitors"                                                     */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void DoLimitActiveIntervalsConstraintMonitorsForResourceAndOffset(       */
/*    KHE_SOLN soln, KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c,                */
/*    KHE_RESOURCE_IN_SOLN rs, int offset)                                   */
/*                                                                           */
/*  Make and attach the monitor for c that monitors r at offset.             */
/*                                                                           */
/*****************************************************************************/

static void DoLimitActiveIntervalsConstraintMonitorsForResourceAndOffset(
  KHE_SOLN soln, KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c,
  KHE_RESOURCE_IN_SOLN rs, int offset)
{
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR m;
  m = KheLimitActiveIntervalsMonitorMake(rs, offset, c);
  KheMonitorAttachToSoln((KHE_MONITOR) m);
  KheGroupMonitorAddChildMonitor((KHE_GROUP_MONITOR) soln, (KHE_MONITOR) m);
}


/*****************************************************************************/
/*                                                                           */
/*  void DoLimitActiveIntervalsConstraintMonitorsForResource(                */
/*    KHE_SOLN soln, KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r)*/
/*                                                                           */
/*  Make and attach the monitors for this constraint that monitor r.         */
/*                                                                           */
/*****************************************************************************/

static void DoLimitActiveIntervalsConstraintMonitorsForResource(
  KHE_SOLN soln, KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_RESOURCE r)
{
  KHE_RESOURCE_IN_SOLN rs;  int i, count;
  rs = KheSolnResourceInSoln(soln, r);
  count = KheLimitActiveIntervalsConstraintAppliesToOffsetCount(c);
  for( i = 0;  i < count;  i++ )
    DoLimitActiveIntervalsConstraintMonitorsForResourceAndOffset(soln, c, rs,
      KheLimitActiveIntervalsConstraintAppliesToOffset(c, i));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheLimitActiveIntervalsConstraintMakeAndAttachMonitors(             */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_SOLN soln)                */
/*                                                                           */
/*  Make and attach the monitors for this constraint.                        */
/*                                                                           */
/*****************************************************************************/

void KheLimitActiveIntervalsConstraintMakeAndAttachMonitors(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KHE_SOLN soln)
{
  int i, j;  KHE_RESOURCE_GROUP rg;  KHE_RESOURCE r;
  HaArrayForEach(c->resource_groups, rg, i)
  {
    for( j = 0;  j < KheResourceGroupResourceCount(rg);  j++ )
    {
      r = KheResourceGroupResource(rg, j);
      DoLimitActiveIntervalsConstraintMonitorsForResource(soln, c, r);
    }
  }
  HaArrayForEach(c->resources, r, i)
    DoLimitActiveIntervalsConstraintMonitorsForResource(soln, c, r);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "reading and writing"                                          */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitActiveIntervalsConstraintMakeFromKml(KML_ELT cons_elt,      */
/*    KHE_INSTANCE ins, KML_ERROR *ke)                                       */
/*                                                                           */
/*  Make a limit active intervals constraint based on cons_elt; add to ins.  */
/*                                                                           */
/*****************************************************************************/

bool KheLimitActiveIntervalsConstraintMakeFromKml(KML_ELT cons_elt,
  KHE_INSTANCE ins, KML_ERROR *ke)
{
  char *id, *name, *ref;  bool reqd;  int i, wt, val, minimum, maximum;
  KHE_COST_FUNCTION cf; KML_ELT elt, history_elt;  KHE_RESOURCE r;
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT res;  KHE_TIME_GROUP applies_to_tg;
  HA_ARENA a;

  /* verify cons_elt and get the common fields */
  a = KheInstanceArena(ins);
  HnAssert(KheInstanceModel(ins) == KHE_MODEL_EMPLOYEE_SCHEDULE,
    "KheLimitActiveIntervalsConstraintMakeFromKml: not employee secheduling");
  if( !KmlCheck(cons_elt, "Id : $Name $Required #Weight $CostFunction AppliesTo"
      " +ResourceHistory +AppliesToTimeGroup TimeGroups #Minimum #Maximum",ke) )
    return false;
  if( !KheConstraintCheckKml(cons_elt, &id, &name, &reqd, &wt, &cf, ke, a) )
    return false;

  /* get minimum and maximum */
  KmlContainsChild(cons_elt, "Minimum", &elt);
  sscanf(KmlText(elt), "%d", &minimum);
  KmlContainsChild(cons_elt, "Maximum", &elt);
  sscanf(KmlText(elt), "%d", &maximum);

  /* get optional applies_to_tg */
  applies_to_tg = NULL;  /* default value */
  if( KmlContainsChild(cons_elt, "AppliesToTimeGroup", &elt) )
  {
    if( !KmlCheck(elt, "Reference", ke) )
      return false;
    ref = KmlAttributeValue(elt, 0);
    if( !KheInstanceRetrieveTimeGroup(ins, ref, &applies_to_tg) )
      return KmlError(ke, a, KmlLineNum(elt), KmlColNum(elt),
	  "<AppliesToTimeGroup> Reference \"%s\" unknown", ref);
  }

  /* build and insert the constraint object */
  if( !KheLimitActiveIntervalsConstraintMake(ins, id, name, reqd, wt, cf,
        applies_to_tg, minimum, maximum, &res) )
    return KmlError(ke, a, KmlLineNum(cons_elt), KmlColNum(cons_elt),
      "<LimitActiveIntervalsConstraint> Id \"%s\" used previously", id);

  /* add the resource groups and resources */
  elt = KmlChild(cons_elt, 4);
  if( !KmlCheck(elt, ": +ResourceGroups +Resources", ke) )
    return false;
  if( !KheConstraintAddResourceGroupsFromKml((KHE_CONSTRAINT) res, elt, ke, a) )
    return false;
  if( !KheConstraintAddResourcesFromKml((KHE_CONSTRAINT) res, elt, ke, a) )
    return false;
  if( KheLimitActiveIntervalsResourceCount(res) == 0 )
    return KmlError(ke, a, KmlLineNum(cons_elt), KmlColNum(cons_elt),
      "<LimitActiveIntervalsConstraint> applies to 0 resources");

  /* add the optional resource history */
  if( KmlContainsChild(cons_elt, "ResourceHistory", &history_elt) )
  {
    /* attributes before_dft and after_dft */
    if( !KmlCheck(history_elt, "#before #after : *Resource", ke) )
      return false;
    sscanf(KmlAttributeValue(history_elt, 0), "%d", &val);
    KheLimitActiveIntervalsConstraintAddHistoryBefore(res, val);
    sscanf(KmlAttributeValue(history_elt, 1), "%d", &val);
    KheLimitActiveIntervalsConstraintAddHistoryAfter(res, val);

    /* children */
    for( i = 0;  i < KmlChildCount(history_elt);  i++ )
    {
      elt = KmlChild(history_elt, i);
      if( !KmlCheck(elt, "Reference", ke) )
	return false;
      if( !KheInstanceRetrieveResource(ins, KmlAttributeValue(elt, 0), &r) )
	return KmlError(ke, a, KmlLineNum(elt), KmlColNum(elt),
	  "unknown resource %s in <Resource>", KmlAttributeValue(elt, 0));
      if( KheLimitActiveIntervalsConstraintHistory(res, r, &val) )
	return KmlError(ke, a, KmlLineNum(elt), KmlColNum(elt),
	  "resource %s appears twice in <ResourceHistory>",
	  KmlAttributeValue(elt, 0));
      if( sscanf(KmlText(elt), "%d", &val) != 1 )
	return KmlError(ke, a, KmlLineNum(elt), KmlColNum(elt),
	  "resource %s has a non-integer body", KmlAttributeValue(elt, 0));
      if( val < 0 )
	return KmlError(ke, a, KmlLineNum(elt), KmlColNum(elt),
	  "resource %s has a negative body", KmlAttributeValue(elt, 0));
      KheLimitActiveIntervalsConstraintAddHistory(res, r, val);
    }
  }

  /* time groups */
  if( !KheConstraintAddTimeGroupsFromKml((KHE_CONSTRAINT) res, cons_elt, ke,a) )
    return false;
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheLimitActiveIntervalsConstraintWrite(                             */
/*    KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KML_FILE kf)                  */
/*                                                                           */
/*  Write c to kf.                                                           */
/*                                                                           */
/*****************************************************************************/

void KheLimitActiveIntervalsConstraintWrite(
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT c, KML_FILE kf)
{
  KHE_RESOURCE_GROUP rg;  KHE_RESOURCE r;  int i, val;
  KHE_TIME_GROUP tg;  char buff[50];
  KmlBegin(kf, "LimitActiveIntervalsConstraint");
  HnAssert(c->id != NULL,
    "KheArchiveWrite: Id missing in LimitActiveIntervalsConstraint");
  KmlAttribute(kf, "Id", c->id);
  KheConstraintWriteCommonFields((KHE_CONSTRAINT) c, kf);

  /* <AppliesTo> */
  KmlBegin(kf, "AppliesTo");
  if( HaArrayCount(c->resource_groups) > 0 )
  {
    KmlBegin(kf, "ResourceGroups");
    HaArrayForEach(c->resource_groups, rg, i)
    {
      HnAssert(KheResourceGroupId(rg) != NULL, "KheArchiveWrite:  Id missing "
        "in ResourceGroup referenced from LimitActiveIntervalsConstraint %s",
	c->id);
      KmlEltAttribute(kf, "ResourceGroup", "Reference", KheResourceGroupId(rg));
    }
    KmlEnd(kf, "ResourceGroups");
  }
  if( HaArrayCount(c->resources) > 0 )
  {
    KmlBegin(kf, "Resources");
    HaArrayForEach(c->resources, r, i)
    {
      HnAssert(KheResourceId(r) != NULL, "KheArchiveWrite:  Id missing in "
        "Resource referenced from LimitActiveIntervalsConstraint %s", c->id);
      KmlEltAttribute(kf, "Resource", "Reference", KheResourceId(r));
    }
    KmlEnd(kf, "Resources");
  }
  KmlEnd(kf, "AppliesTo");

  /* <ResourceHistory> */
  if( c->history_before != -1 || c->history_after != -1 ||
      HaArrayCount(c->history) > 0 )
  {
    KmlBegin(kf, "ResourceHistory");
    sprintf(buff, "%d", c->history_before == -1 ? 0 : c->history_before);
    KmlAttribute(kf, "before", buff);
    sprintf(buff, "%d", c->history_after == -1 ? 0 : c->history_after);
    KmlAttribute(kf, "after", buff);
    for( i = 0;  i < HaArrayCount(c->history);  i++ )
    {
      val = HaArray(c->history, i);
      if( val != -1 )
      {
	r = KheInstanceResource(c->instance, i);
	KmlEltAttributeFmtText(kf, "Resource", "Reference", KheResourceId(r),
	  "%d", val);
      }
    }
    KmlEnd(kf, "ResourceHistory");
  }

  /* <AppliesToTimeGroup> */
  if( c->applies_to_tg != NULL )
  {
    KmlBegin(kf, "AppliesToTimeGroup");
    KmlAttribute(kf, "Reference", KheTimeGroupId(c->applies_to_tg));
    KmlEnd(kf, "AppliesToTimeGroup");
  }

  /* <TimeGroups> */
  KmlBegin(kf, "TimeGroups");
  HaArrayForEach(c->time_groups, tg, i)
  {
    HnAssert(KheTimeGroupId(tg) != NULL, "KheArchiveWrite:  Id missing in "
      "TimeGroup referenced from LimitActiveIntervalsConstraint %s", c->id);
    KmlBegin(kf, "TimeGroup");
    KmlAttribute(kf, "Reference", KheTimeGroupId(tg));
    if( HaArray(c->polarities, i) == KHE_NEGATIVE )
      KmlAttribute(kf, "Polarity", "negative");
    KmlEnd(kf, "TimeGroup");
  }
  KmlEnd(kf, "TimeGroups");

  /* <Minimum> and <Maximum> */
  KmlEltFmtText(kf, "Minimum", "%d", c->minimum);
  KmlEltFmtText(kf, "Maximum", "%d", c->maximum);
  KmlEnd(kf, "LimitActiveIntervalsConstraint");
}
