
/*****************************************************************************/
/*                                                                           */
/*  THE KTS TIMETABLING SYSTEM                                               */
/*  COPYRIGHT (C) 2004, 2008 Jeffrey H. Kingston                             */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  FILE:         kml_read.c                                                 */
/*  MODULE:       XML reading                                                */
/*                                                                           */
/*  This module parses an XML file using this grammar ($ denotes simplified) */
/*                                                                           */
/*    <document>  ::=  <prolog> <element> { <misc> }                         */
/*    <prolog>    ::=  { <xmldecl> | <dtdecl> | <misc> }                  $  */
/*    <xmldecl>   ::=  <pi>                                               $  */
/*    <dtdecl>    ::=  "<!" ... ">"            (... is anything but ">")  $  */
/*    <misc>      ::=  <comment> | <pi> | <s>                                */
/*    <comment>   ::=  "<!--" ... "-->"       (... is anything but "--")  $  */
/*    <pi>        ::=  "<?" ... "?>"          (... is anything but "?>")  $  */
/*    <s>         ::=  <os> { <os> }                                         */
/*    <os>        ::=  " "  |  "\t"  |  "\r"  |  "\n"                        */
/*                                                                           */
/*    <element>   ::=  <emptyelt>  |  <stag> <content> <etag>                */
/*    <emptyelt>  ::=  "<" <name> { <s> <attribute> } [ <s> ] "/>"           */
/*    <stag>      ::=  "<" <name> { <s> <attribute> } [ <s> ] ">"            */
/*    <etag>      ::=  "</" <name> [ <s> ] ">"                               */
/*    <attribute> ::=  <name> [ <s> ] "=" [ <s> ] <attvalue>                 */
/*    <attvalue>  ::=  "\"" { <any1> | <reference> } "\""                    */
/*    <attvalue>  ::=  "'"  { <any2> | <reference> } "'"                     */
/*    <any1>      ::=  (any character except "<", "&", and "\"")             */
/*    <any2>      ::=  (any character except "<", "&", and "'")              */
/*    <reference> ::=  "&" <name> ";"                                     $  */
/*                                                                           */
/*    <content>   ::=  [ <chardata> ] { <cintern> [ <chardata> ] }           */
/*    <chardata>  ::=  sequence of chars not containing "<", "&", or "]]>"   */
/*    <cintern>   ::=  <element> | <reference> | <cdsect> | <pi> | <comment> */
/*    <cdsect>    ::=  "<![CDATA[" ... "]]>"     (... is anything but "]]>"  */
/*                                                                           */
/*  Implementation note.  The implementation uses a recursive descent        */
/*  parser.  There is no lexical analyser; instead the parser operates       */
/*  directly on characters, with one character of lookahead.                 */
/*                                                                           */
/*****************************************************************************/
#include <setjmp.h>
#include "kml.h"
#include "howard_a.h"
#include "howard_n.h"

#define DEBUG1 0
#define DEBUG2 0
#define DEBUG3 0

#define is_space(ch) ((ch)==' ' || (ch)=='\t' || (ch)=='\n' || (ch)=='\r')
#define is_letter(ch) (((ch)>='A' && (ch)<='Z') || ((ch)>='a' && (ch)<='z'))
#define is_digit(ch) ((ch) >= '0' && (ch) <= '9')
#define is_other_name(ch) ((ch)=='_' || (ch)==':' || (ch)=='-' || (ch)=='.')


/*****************************************************************************/
/*                                                                           */
/*  KML_READ_INFO - read info passed through the parser to the user          */
/*                                                                           */
/*****************************************************************************/

struct kml_read_info_rec {

  /* input file */
  FILE			*fp;			/* file being read           */
  FILE			*echo_fp;		/* echo to file, if present  */
  int			line_num;		/* line number of curr_ch    */
  int			col_num;		/* column number of curr_ch  */
  int			curr_ch;		/* first unconsumed char     */

  /* user callbacks */
  KML_ELT_FN		elt_fn;			/* call after each elt       */
  void			*impl;			/* pass to elt_fn            */
  HA_ARENA		arena;			/* pass to elt_fn            */
  int			max_depth;		/* max depth of callbacks    */
  int			curr_depth;		/* current depth of parse    */

  /* error handling */
  KML_ERROR		ke;			/* error object              */
  jmp_buf		jmp_env;		/* early return              */
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "lexical analyser" (private)                                   */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  int KmlCurrChar(KML_READ_INFO ri)                                        */
/*                                                                           */
/*  Return the next unconsumed character, possible EOF.                      */
/*                                                                           */
/*****************************************************************************/

static int KmlCurrChar(KML_READ_INFO ri)
{
  return ri->curr_ch;
}


/*****************************************************************************/
/*                                                                           */
/*  void KmlNextChar(KML_READ_INFO ri)                                       */
/*                                                                           */
/*  Read the next character from fp, and increment ri's line number and      */
/*  column number attributes as appropriate.  The result is not '\r'.        */
/*                                                                           */
/*    <ch> ::= "\n" | "\n\r" | "\r" | "\r\n" | ...                           */
/*                                                                           */
/*  where ... stands for any other single character.  The sequences of       */
/*  two newline characters read here are converted into single characters.   */
/*                                                                           */
/*****************************************************************************/

static void KmlNextChar(KML_READ_INFO ri)
{
  int ch;
  ch = getc(ri->fp);
  if( ch == '\n' )
  {
    /* "\n" | "\n\r" */
    ch = getc(ri->fp);
    if( ch != '\r' )
      ungetc(ch, ri->fp);
    ri->line_num++, ri->col_num = 0;
    ri->curr_ch = '\n';
  }
  else if( ch == '\r' )
  {
    /* "\r" | "\r\n" */
    ch = getc(ri->fp);
    if( ch != '\n' )
      ungetc(ch, ri->fp);
    ri->line_num++, ri->col_num = 0;
    ri->curr_ch = '\n';
  }
  else
  {
    /* ... */
    ri->col_num++;
    ri->curr_ch = ch;
  }
  if( ri->echo_fp != NULL && ri->curr_ch != EOF )
    putc(ri->curr_ch, ri->echo_fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "character-skipping functions" (private)                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KmlSkipChars(KML_READ_INFO ri, char *start_str, char *end_str,      */
/*    HA_ARENA a)                                                            */
/*                                                                           */
/*  Skip chars until KmlCurrChar is the first character after end_str:       */
/*                                                                           */
/*    <start_str> ... <end_str>    (where ... is anything except <end_str>)  */
/*                                                                           */
/*****************************************************************************/

static void KmlSkipChars(KML_READ_INFO ri, char *start_str, char *end_str,
  HA_ARENA a)
{
  char buff[5];  int i, len, line_num, col_num;
  len = strlen(end_str);
  HnAssert(len > 0 && len < 5, "KmlSkipChars internal error");
  for( i = 0;  i <= len;  i++ )
    buff[i] = '\0';
  line_num = ri->line_num, col_num = ri->col_num;
  for( ;; )
  {
    /* shift buff and add current char to it */
    for( i = 1;  i < len;  i++ )
      buff[i-1] = buff[i];
    buff[i-1] = KmlCurrChar(ri);

    /* move to next char */
    KmlNextChar(ri);

    /* exit if match */
    if( strcmp(end_str, buff) == 0 )
      return;

    /* fail if current char is EOF */
    if( KmlCurrChar(ri) == EOF )
      KmlReadFail(ri, KmlErrorMake(a, line_num, col_num,
	"\"%s\" has no matching \"%s\"", start_str, end_str));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KmlSkipZeroOrMoreSpaces(KML_READ_INFO ri)                           */
/*                                                                           */
/*  Skip zero or more white space characters.                                */
/*                                                                           */
/*      [ <s> ]  aka  { <os> }                                               */
/*                                                                           */
/*****************************************************************************/

static void KmlSkipZeroOrMoreSpaces(KML_READ_INFO ri)
{
  while( is_space(KmlCurrChar(ri)) )
    KmlNextChar(ri);
}


/*****************************************************************************/
/*                                                                           */
/*  void KmlSkipOneOrMoreSpaces(KML_READ_INFO ri)                            */
/*                                                                           */
/*  Skip one or more white space characters.                                 */
/*                                                                           */
/*****************************************************************************/

/* *** unused owing to various grammar transformations
static void KmlSkipOneOrMoreSpaces(KML_READ_INFO ri)
{
  if( !is_space(KmlCurrChar(ri)) )
    KmlReadFail(ri, KmlErrorMake(ri->line_num, ri->col_num,
      "expected white space character here"));
  do
  {
    KmlNextChar(ri);
  }
  while( is_space(KmlCurrChar(ri)) );
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  Submodule "names and references" (private)                               */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KmlCharIsNameChar(int ch)                                           */
/*                                                                           */
/*  Return true if ch may occur within a name.  This is only a rough stab    */
/*  at the XML specification.                                                */
/*                                                                           */
/*****************************************************************************/

static bool KmlCharIsNameChar(int ch)
{
  return is_letter(ch) || is_digit(ch) || is_other_name(ch);
}


/*****************************************************************************/
/*                                                                           */
/*  void KmlParseName(KML_READ_INFO ri, char buff[501], HA_ARENA a)          */
/*                                                                           */
/*  Parse a name.  The result is in buff.                                    */
/*                                                                           */
/*****************************************************************************/

static void KmlParseName(KML_READ_INFO ri, char buff[501], HA_ARENA a)
{
  int i;
  for( i = 0;  i < 500 && KmlCharIsNameChar(KmlCurrChar(ri));  i++ )
  {
    buff[i] = KmlCurrChar(ri);
    KmlNextChar(ri);
  }
  buff[i] = '\0';
  if( i >= 500 )
    KmlReadFail(ri, KmlErrorMake(a, ri->line_num, ri->col_num,
      "name %s... too long", buff));
}


/*****************************************************************************/
/*                                                                           */
/*  char KmlParseReference(KML_READ_INFO ri, HA_ARENA a)                     */
/*                                                                           */
/*  Parse one reference, returning the character it refers to.               */
/*                                                                           */
/*    <reference> ::=  "&" <name> ";"                                        */
/*                                                                           */
/*****************************************************************************/

static char KmlParseReference(KML_READ_INFO ri, HA_ARENA a)
{
  char name[501];
  KmlNextChar(ri);  /* skip opening & */
  KmlParseName(ri, name, a);
  if( KmlCurrChar(ri) != ';' )
    KmlReadFail(ri, KmlErrorMake(a, ri->line_num, ri->col_num,
      "missing ; after &%s", name));
  KmlNextChar(ri);
  if( strcmp(name, "lt") == 0 )
    return '<';
  else if( strcmp(name, "gt") == 0 )
    return '>';
  else if( strcmp(name, "amp") == 0 )
    return '&';
  else if( strcmp(name, "apos") == 0 )
    return '\'';
  else if( strcmp(name, "quot") == 0 )
    return '"';
  else
  {
    KmlReadFail(ri, KmlErrorMake(a, ri->line_num, ri->col_num,
      "reference &%s; unknown", name));
    return '\0';  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "main part of parser" (private)                                */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KmlParseProlog(KML_READ_INFO ri, HA_ARENA a)                        */
/*                                                                           */
/*  Parse the prolog.  When this function returns, the first "<" of the      */
/*  root element has been read.  Officially, the syntax is                   */
/*                                                                           */
/*    <prolog>    ::=  [ <xmldecl> ] { <misc> } [ <dtdecl> { <misc> } ]      */
/*                                                                           */
/*  but the syntax actually parsed here is the more general and easier       */
/*                                                                           */
/*    <prolog>    ::= { <xmldecl> | <dtdecl> | <comment> | <pi> | <s> } "<"  */
/*                                                                           */
/*****************************************************************************/

static void KmlParseProlog(KML_READ_INFO ri, HA_ARENA a)
{
  if( DEBUG1 )
    fprintf(stderr, "[ KmlParseProlog(ri)\n");
  for( ;; ) switch( KmlCurrChar(ri) )
  {
    case '<':

      /* <xmldecl>, <dtdecl>, <comment>, or <pi> */
      KmlNextChar(ri);
      if( KmlCurrChar(ri) == '!' )
      {
	/* <dtdecl> or <comment> */
	KmlNextChar(ri);
	if( KmlCurrChar(ri) == '-' )
	{
	  /* <comment> */
	  KmlSkipChars(ri, "<!-", "-->", a);
	}
	else
	{
	  /* <dtdecl> */
	  KmlSkipChars(ri, "<!", ">", a);
	}
      }
      else if( KmlCurrChar(ri) == '?' )
      {
	/* <xmldecl> or <pi> */
	KmlSkipChars(ri, "<?", "?>", a);
      }
      else
      {
	/* the only return; curr char is char following "<" */
	if( DEBUG1 )
	  fprintf(stderr, "] KmlParseProlog\n");
	return;
      }
      break;

    /* <s> */
    case '\n':
    case '\t':
    case ' ':

      KmlNextChar(ri);
      break;

    default:

      if( KmlCurrChar(ri) == EOF )
	KmlReadFail(ri, KmlErrorMake(a, ri->line_num, ri->col_num,
	  "unexpected end of file in prolog"));
      else
	KmlReadFail(ri, KmlErrorMake(a, ri->line_num, ri->col_num,
	  "unexpected character '%c' in prolog", KmlCurrChar(ri)));
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KmlParseAttributeValue(KML_READ_INFO ri,                            */
/*    char *attribute_name, char buff[501], HA_ARENA a)                      */
/*                                                                           */
/*  Parse one attribute value, including the opening and closing delimiter,  */
/*  which is KmlCurrChar(ri).  The result is in buff.                        */
/*                                                                           */
/*    <attvalue>  ::=  "\"" { <any1> | <reference> } "\""                    */
/*    <attvalue>  ::=  "'" { <any2> | <reference> } "'"                      */
/*    <any1>      ::=  (any character except "<", "&", and "\"")             */
/*    <any2>      ::=  (any character except "<", "&", and "'")              */
/*                                                                           */
/*****************************************************************************/

static void KmlParseAttributeValue(KML_READ_INFO ri,
  char *attribute_name, char buff[501], HA_ARENA a)
{
  int i, delim_ch, line_num, col_num;
  delim_ch = KmlCurrChar(ri);
  line_num = ri->line_num, col_num = ri->col_num;
  KmlNextChar(ri);  /* skip opening delimiter */
  for( i = 0;  KmlCurrChar(ri) != delim_ch;  i++ )
  {
    if( i >= 500 )
      KmlReadFail(ri, KmlErrorMake(a, line_num, col_num,
	"value of attribute %s is too long", attribute_name));
    else if( KmlCurrChar(ri) == '<' )
      KmlReadFail(ri, KmlErrorMake(a, line_num, col_num,
	"'<' character within value of attribute %s", attribute_name));
    else if( KmlCurrChar(ri) == EOF )
      KmlReadFail(ri, KmlErrorMake(a, line_num, col_num,
	"end of input within value of attribute %s", attribute_name));
    else if( KmlCurrChar(ri) == '&' )
      buff[i] = KmlParseReference(ri, a);
    else
    {
      buff[i] = KmlCurrChar(ri);
      KmlNextChar(ri);
    }
  }
  buff[i] = '\0';
  KmlNextChar(ri);  /* skip closing delimiter */
}


/*****************************************************************************/
/*                                                                           */
/*  void KmlParseAttribute(KML_READ_INFO ri, KML_ELT prnt, HA_ARENA a)       */
/*                                                                           */
/*  Parse an attribute and add it to prnt.  The syntax is                    */
/*                                                                           */
/*    <attribute> ::=  <name> [ <s> ] "=" [ <s> ] <attvalue>                 */
/*                                                                           */
/*****************************************************************************/

static void KmlParseAttribute(KML_READ_INFO ri, KML_ELT prnt, HA_ARENA a)
{
  char name[501], value[501];
  KmlParseName(ri, name, a);
  KmlSkipZeroOrMoreSpaces(ri);
  if( KmlCurrChar(ri) != '=' )
    KmlReadFail(ri, KmlErrorMake(a, ri->line_num, ri->col_num,
      "expected = after attribute name %s", name));
  KmlNextChar(ri);
  KmlSkipZeroOrMoreSpaces(ri);
  if( KmlCurrChar(ri) != '"' && KmlCurrChar(ri) != '\'' )
    KmlReadFail(ri, KmlErrorMake(a, ri->line_num, ri->col_num,
      "expected ' or \" at the start of value of attribute %s", name));
  KmlParseAttributeValue(ri, name, value, a);
  KmlAddAttribute(prnt, name, value);
}


/*****************************************************************************/
/*                                                                           */
/*  void KmlParseContent(KML_READ_INFO ri, KML_ELT prnt, HA_ARENA a)         */
/*                                                                           */
/*  Parse content and add it to prnt.  The syntax is                         */
/*                                                                           */
/*    <content>   ::=  [ <chardata> ] { <cintern> [ <chardata> ] }           */
/*    <chardata>  ::=  sequence of chars not containing "<", "&", or "]]>"   */
/*    <cintern>   ::=  <element> | <reference> | <cdsect> | <pi> | <comment> */
/*    <cdsect>    ::=  "<![CDATA[" ... "]]>"     (... is anything but "]]>"  */
/*                                                                           */
/*  Actually this function parses the equivalent grammar                     */
/*                                                                           */
/*    <content>   ::=  { <reference> | <cstuff> | <chardata> }               */
/*    <cstuff>    ::=  <cdsect> | <comment> | <pi> | <element>               */
/*                                                                           */
/*  When it returns it has already consumed the initial "<" of the <etag>.   */
/*                                                                           */
/*****************************************************************************/
static KML_ELT KmlParseElement(KML_READ_INFO ri, KML_ELT prnt, HA_ARENA a);

static void KmlParseContent(KML_READ_INFO ri, KML_ELT prnt, HA_ARENA a)
{
  int i;  char buff[501];
  if( DEBUG3 )
    fprintf(stderr, "%*s[ ParseContent(...)\n", ri->curr_depth * 2, "");
  i = 0;
  for( ;; )
  {
    switch( KmlCurrChar(ri) )
    {
      case '&':

	/* <reference> */
	if( DEBUG3 )
	  fprintf(stderr, "%*s  ParseContent (2)\n", ri->curr_depth * 2, "");
        buff[i++] = KmlParseReference(ri, a);
	if( i >= 500 )
	{
	  buff[i] = '\0';
	  KmlAddText(prnt, buff);
	  i = 0;
	}
	break;

      case '<':

	/* <cstuff> or start of <etag> */
	if( DEBUG3 )
	  fprintf(stderr, "%*s  ParseContent (3)\n", ri->curr_depth * 2, "");
	KmlNextChar(ri);
	switch( KmlCurrChar(ri) )
	{
	  case '/':

	    /* <etag> ::=  "</" <name> [ <s> ] ">" */
	    if( DEBUG3 )
	      fprintf(stderr, "%*s  ParseContent (4)\n", ri->curr_depth*2, "");
	    if( i > 0 )
	    {
	      buff[i] = '\0';
	      KmlAddText(prnt, buff);
	    }
	    if( DEBUG3 )
	      fprintf(stderr, "%*s] ParseContent returning\n",
		ri->curr_depth * 2, "");
	    return;  /* only way out */

	  case '!':

	    /* <cdsect> | <comment> */
	    if( DEBUG3 )
	      fprintf(stderr, "%*s  ParseContent (5)\n", ri->curr_depth*2, "");
	    KmlNextChar(ri);
	    if( KmlCurrChar(ri) == '-' )
	    {
	      /* <comment> */
	      KmlSkipChars(ri, "<!-", "-->", a);
	    }
	    else
	    {
	      /* <cdsect> */
	      KmlSkipChars(ri, "<!", "]]>", a);
	    }
	    break;

	  case '?':

	    /* <pi> */
	    if( DEBUG3 )
	      fprintf(stderr, "%*s  ParseContent (6)\n", ri->curr_depth*2, "");
            KmlSkipChars(ri, "<?", "?>", a);
	    break;

	  default:

	    /* <element> */
	    if( DEBUG3 )
	      fprintf(stderr, "%*s  ParseContent (7)\n", ri->curr_depth*2, "");
            KmlParseElement(ri, prnt, a);
	    break;
	}
	break;

      case EOF:

	/* unexpected end of file */
        KmlReadFail(ri, KmlErrorMake(a, ri->line_num, ri->col_num,
          "unexpected end of file (within %s element at line %d, col %d)",
	  KmlLabel(prnt), KmlLineNum(prnt), KmlLineNum(prnt)));
	break;

      default:

        /* <chardata> */
	buff[i++] = KmlCurrChar(ri);
	if( i >= 500 )
	{
	  buff[i] = '\0';
	  KmlAddText(prnt, buff);
	  i = 0;
	}
	KmlNextChar(ri);
	break;
    }
  }
  HnAbort("KmlParseContent internal error");
}


/*****************************************************************************/
/*                                                                           */
/*  KML_ELT KmlParseElement(KML_READ_INFO ri, KML_ELT prnt, HA_ARENA a)      */
/*                                                                           */
/*  Parse an <element> according to syntax                                   */
/*                                                                           */
/*    <element>   ::=  <emptyelt>  |  <stag> <content> <etag>                */
/*    <emptyelt>  ::=  "<" <name> { <s> <attribute> } [ <s> ] "/>"           */
/*    <stag>      ::=  "<" <name> { <s> <attribute> } [ <s> ] ">"            */
/*    <etag>      ::=  "</" <name> [ <s> ] ">"                               */
/*                                                                           */
/*  Actually the initial "<" has already been read when this is called,      */
/*  and there are LL difficulties in the grammar as given.  So the grammar   */
/*  actually parsed is                                                       */
/*                                                                           */
/*    <element>   ::=  <elem_init> <elem_rest>                               */
/*    <elem_init> ::=  <name> [ <s> ] { <attribute> [ <s> ] }                */
/*    <elem_rest> ::=  "/>"                                                  */
/*    <elem_rest> ::=  ">" <content> <etag>                                  */
/*    <etag>      ::=  "</" <name> [ <s> ] ">"                               */
/*                                                                           */
/*  The resulting KML_ELT is returned, and also added to prnt if non-NULL.   */
/*                                                                           */
/*****************************************************************************/

static KML_ELT KmlParseElement(KML_READ_INFO ri, KML_ELT prnt, HA_ARENA a)
{
  KML_ELT res;  char name[501], end_name[501];
  int line_num, col_num, end_line_num, end_col_num;

  /* <name> */
  line_num = ri->line_num;  col_num = ri->col_num - 1;
  KmlParseName(ri, name, a);
  res = KmlMakeElt(line_num, col_num, name, a);
  if( prnt != NULL )
    KmlAddChild(prnt, res);
  ri->curr_depth++;
  if( DEBUG2 )
    fprintf(stderr, "%*s[ KmlParseElement(), line %d col %d name %s depth %d\n",
      ri->curr_depth * 2, "", ri->line_num, ri->col_num, name, ri->curr_depth);

  /* [ <s> ] { <attribute> [ <s> ] } */
  KmlSkipZeroOrMoreSpaces(ri);
  while( KmlCurrChar(ri) != '>' && KmlCurrChar(ri) != '/' )
  {
    KmlParseAttribute(ri, res, a);
    KmlSkipZeroOrMoreSpaces(ri);
  }
  if( DEBUG2 )
    fprintf(stderr, "%*s  KmlParseElement (2)\n", ri->curr_depth * 2, "");

  /* <elem_rest> */
  if( KmlCurrChar(ri) == '>' )
  {
    /* ">" */
    KmlNextChar(ri);

    /* <content> */
    if( DEBUG2 )
      fprintf(stderr, "%*s  KmlParseElement (3)\n", ri->curr_depth * 2, "");
    KmlParseContent(ri, res, a);
    if( DEBUG2 )
      fprintf(stderr, "%*s  KmlParseElement (4)\n", ri->curr_depth * 2, "");

    /* <etag> ::= "</" <name> [ <s> ] ">"   (except "<" is already read) */
    end_line_num = ri->line_num;  end_col_num = ri->col_num - 1;
    /* ***
    if( KmlCurrChar(ri) != '<' )
      KmlReadFail(ri, KmlErrorMake(end_line_num, end_col_num,
	"expected content or </%s> here", name));
    KmlNextChar(ri);
    *** */
    if( KmlCurrChar(ri) != '/' )
      KmlReadFail(ri, KmlErrorMake(a, ri->line_num, ri->col_num,
	"expected / of </%s> here", name));
    KmlNextChar(ri);
    KmlParseName(ri, end_name, a);
    if( strcmp(end_name, name) != 0 )
      KmlReadFail(ri, KmlErrorMake(a, end_line_num, end_col_num,
	"</%s> failed to match <%s> at line %d, column %d", end_name,
	name, line_num, col_num));
    KmlSkipZeroOrMoreSpaces(ri);
    if( KmlCurrChar(ri) != '>' )
      KmlReadFail(ri, KmlErrorMake(a, ri->line_num, ri->col_num,
	"expected > of </%s> here", name));
    KmlNextChar(ri);
  }
  else /* KmlCurrChar(ri) == '/' */
  {
    KmlNextChar(ri);
    if( KmlCurrChar(ri) != '>' )
      KmlReadFail(ri, KmlErrorMake(a, ri->line_num, ri->col_num,
	"/ not followed by >"));
    KmlNextChar(ri);
  }
  if( ri->curr_depth <= ri->max_depth )
    ri->elt_fn(res, ri);
  if( DEBUG2 )
    fprintf(stderr, "%*s] KmlParseElement()\n", ri->curr_depth * 2, "");
  ri->curr_depth--;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KmlParseDocument(FILE *fp, KML_READ_INFO ri)                        */
/*                                                                           */
/*  Parse an XML document according to the grammar for <document> above.     */
/*                                                                           */
/*    <document>  ::=  <prolog> <element> { <misc> }                         */
/*    <misc>      ::=  <comment> | <pi> | <s>                                */
/*                                                                           */
/*****************************************************************************/

static KML_ELT KmlParseDocument(KML_READ_INFO ri, HA_ARENA a)
{
  KML_ELT res;
  if( DEBUG1 )
    fprintf(stderr, "[ KmlParseDocument(ri)\n");

  /* <prolog> */
  KmlParseProlog(ri, a);

  /* <element> */
  res = KmlParseElement(ri, NULL, a);

  /* { <comment> | <pi> | <s> } */
  switch( KmlCurrChar(ri) )
  {
    case '<':

      /* <comment> | <pi> */
      KmlNextChar(ri);
      switch( KmlCurrChar(ri) )
      {
	case '!':

	  /* <comment> */
          KmlNextChar(ri);
	  if( KmlCurrChar(ri) != '-' )
	    KmlReadFail(ri, KmlErrorMake(a, ri->line_num, ri->col_num,
	      "in epilogue, <! not followed by -"));
          KmlSkipChars(ri, "<!-", "-->", a);
	  break;

	case '?':

	  /* <pi> */
          KmlSkipChars(ri, "<?", "?>", a);
	  break;

	default:

	  KmlReadFail(ri, KmlErrorMake(a, ri->line_num, ri->col_num,
	    "in epilogue, < not followed by ! or ?"));
	  break;
      }
      break;

    case ' ':
    case '\t':
    case '\n':

      /* <s> */
      break;

    case EOF:

      /* normal exit */
      return res;

    default:
      
      /* following non-XML, preserved for whoever wants it */
      ungetc(KmlCurrChar(ri), ri->fp);
  }
  if( DEBUG1 )
    fprintf(stderr, "] KmlParseDocument returning\n");
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "public interface"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KmlReadFile(FILE *fp, KML_ELT *res, KML_ERROR *ke, FILE *echo_fp,   */
/*    HA_ARENA a)                                                            */
/*                                                                           */
/*  Read an XML element from fp.  If successful, set *res to the element     */
/*  and return true.  Otherwise set *ke to an error object describing the    */
/*  first problem encountered, and return false.                             */
/*                                                                           */
/*****************************************************************************/

bool KmlReadFile(FILE *fp, KML_ELT *res, KML_ERROR *ke, FILE *echo_fp,
  HA_ARENA a)
{
  return KmlReadFileIncremental(fp, res, ke, echo_fp, NULL, NULL, -1, a);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KmlReadString(char *str, char *stop_str, KML_ELT *res,KML_ERROR *ke)*/
/*                                                                           */
/*  Like KmlReadFile just above, but reading from a string, not from a file. */
/*                                                                           */
/*****************************************************************************/

bool KmlReadString(char *str, KML_ELT *res, KML_ERROR *ke, HA_ARENA a)
{
  /* still to do */
  HnAbort("KmlReadString still to do");
  return false;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KmlReadFileIncremental(FILE *fp, KML_ELT *res, KML_ERROR *ke,       */
/*    FILE *echo_fp, KML_ELT_FN elt_fn, void *impl, int max_depth,           */
/*    HA_ARENA a)                                                            */
/*                                                                           */
/*  Incrementally read fp, calling back after each element of depth at       */
/*  most max_depth.                                                          */
/*                                                                           */
/*****************************************************************************/

bool KmlReadFileIncremental(FILE *fp, KML_ELT *res, KML_ERROR *ke,
  FILE *echo_fp, KML_ELT_FN elt_fn, void *impl, int max_depth, HA_ARENA a)
{
  struct kml_read_info_rec ri_rec;
  if( DEBUG1 )
    fprintf(stderr, "[ KmlReadFileIncremental(...)\n");

  /* initialize the user data record - input file */
  ri_rec.fp = fp;
  ri_rec.echo_fp = echo_fp;
  ri_rec.line_num = 1;
  ri_rec.col_num = 0;
  KmlNextChar(&ri_rec);

  /* initialize the user data record - user callbacks */
  ri_rec.elt_fn = elt_fn;
  ri_rec.impl = impl;
  ri_rec.max_depth = max_depth;
  ri_rec.arena = a;
  ri_rec.curr_depth = -1;

  /* initialize the user data record - error handling */
  ri_rec.ke = NULL;

  /* find a document within fp - or not */
  if( setjmp(ri_rec.jmp_env) == 0 )
  {
    *res = KmlParseDocument(&ri_rec, a);
    *ke = NULL;
    if( DEBUG1 )
      fprintf(stderr, "] KmlReadFileIncremental returning true\n");
    return true;
  }
  else
  {
    *res = NULL;
    *ke = ri_rec.ke;
    if( DEBUG1 )
      fprintf(stderr, "] KmlReadFileIncremental returning false\n");
    return false;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void *KmlReadImpl(KML_READ_INFO ri)                                      */
/*                                                                           */
/*  Return the impl parameter of KmlReadFileIncremental.                     */
/*                                                                           */
/*****************************************************************************/

void *KmlReadImpl(KML_READ_INFO ri)
{
  return ri->impl;
}


/*****************************************************************************/
/*                                                                           */
/*  HA_ARENA KmlReadArena(KML_READ_INFO ri)                                  */
/*                                                                           */
/*  Return the arena parameter of KmlReadFileIncremental.                    */
/*                                                                           */
/*****************************************************************************/

HA_ARENA KmlReadArena(KML_READ_INFO ri)
{
  return ri->arena;
}


/*****************************************************************************/
/*                                                                           */
/*  int KmlReadMaxDepth(KML_READ_INFO ri)                                    */
/*                                                                           */
/*  Return the max_depth parameter of KmlReadFileIncremental.                */
/*                                                                           */
/*****************************************************************************/

int KmlReadMaxDepth(KML_READ_INFO ri)
{
  return ri->max_depth;
}


/*****************************************************************************/
/*                                                                           */
/*  int KmlReadCurrDepth(KML_READ_INFO ri)                                   */
/*                                                                           */
/*  Return the current depth of the read.                                    */
/*                                                                           */
/*****************************************************************************/

int KmlReadCurrDepth(KML_READ_INFO ri)
{
  return ri->curr_depth;
}


/*****************************************************************************/
/*                                                                           */
/*  void KmlReadFail(KML_READ_INFO ri, KML_ERROR ke)                         */
/*                                                                           */
/*  Make an immediate fail exit from the current incremental read, with      */
/*  the given ke.                                                            */
/*                                                                           */
/*****************************************************************************/

void KmlReadFail(KML_READ_INFO ri, KML_ERROR ke)
{
  ri->ke = ke;
  longjmp(ri->jmp_env, 1);
}
