
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sr_group_by_frame.c                                    */
/*  DESCRIPTION:  Group by frame                                             */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"
#include <limits.h>

#define DEBUG1 0
#define DEBUG2 0
#define DEBUG3 0
#define DEBUG4 0


/*****************************************************************************/
/*                                                                           */
/*  KHE_FRAMED_MONITOR - a monitor, plus a frame saying what it covers       */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_framed_monitor_rec
{
  KHE_MONITOR		monitor;
  int			first_index;
  int			last_index;
} *KHE_FRAMED_MONITOR;

typedef HA_ARRAY(KHE_FRAMED_MONITOR) ARRAY_KHE_FRAMED_MONITOR;


/*****************************************************************************/
/*                                                                           */
/*  KHE_TIME_SET - a set of times (like a time group, but without baggage)   */
/*                                                                           */
/*****************************************************************************/

typedef HA_ARRAY(KHE_TIME) ARRAY_KHE_TIME;

typedef struct khe_time_set_rec
{
  ARRAY_KHE_TIME		times;
} *KHE_TIME_SET;

typedef HA_ARRAY(KHE_TIME_SET) ARRAY_KHE_TIME_SET;


/*****************************************************************************/
/*                                                                           */
/*  KHE_GROUP_BY_FRAME_SOLVER - solver for group by frame solving            */
/*                                                                           */
/*  NB start_indexes[si] is the smallest i such that                         */
/*  framed_monitors[i]->start_index >= si, or if there is no such            */
/*  index there is no entry for si in start_indexes.                         */
/*                                                                           */
/*  Consequently, the following code will visit all monitors whose           */
/*  frames start at or after si:                                             */
/*                                                                           */
/*  if( si < HaArrayCount(start_indexes) )                                   */
/*    for( i = start_indexes[si];  i < HaArrayCount(framed_monitors);  i++ ) */
/*      visit(framed_monitors[i];                                            */
/*                                                                           */
/*  and the first time the visit comes upon a monitor whose first index      */
/*  is past the interval we are interested in, we can stop.                  */
/*                                                                           */
/*****************************************************************************/

typedef HA_ARRAY(KHE_LIMIT_ACTIVE_INTERVALS_MONITOR)
  ARRAY_KHE_LIMIT_ACTIVE_INTERVALS_MONITOR;

typedef struct khe_group_by_frame_solver_rec {

  /* the solve as a whole */
  HA_ARENA			arena;
  KHE_TASKING			tasking;
  KHE_SOLN			soln;
  bool				resource_invariant;
  KHE_FRAME			common_frame;
  KHE_EVENT_TIMETABLE_MONITOR	etm;
  KHE_RESOURCE			resource;
  bool				soln_changed;
  KHE_TASK_SET			task_set;

  /* phase 1 */
  int				max_days;
  ARRAY_KHE_FRAMED_MONITOR	framed_monitors;
  HA_ARRAY_INT			start_indexes;
  KHE_GROUP_MONITOR		group_monitor;
  int				last_index;
  int				success_count;
  KHE_TIME_SET			success_ts;
  KHE_TIME_SET			curr_ts;

  /* phase 2 */
  ARRAY_KHE_TIME_SET		time_sets;
  ARRAY_KHE_TIME_SET		time_sets_by_time;
  ARRAY_KHE_TIME_SET		time_sets_by_monitor;
  /* ***
  ARRAY_KHE_LIMIT_ACTIVE_INTERVALS_MONITOR limit_active_intervals_monitors;
  *** */
  HA_ARRAY_INT			success_seq;
  HA_ARRAY_INT			curr_seq;

} *KHE_GROUP_BY_FRAME_SOLVER;


/*****************************************************************************/
/*                                                                           */
/*  Submodule "framed monitors"                                              */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_FRAMED_MONITOR KheFramedMonitorMake(KHE_MONITOR m,                   */
/*    int first_index, int last_index)                                       */
/*                                                                           */
/*  Make a new framed monitor object with these attributes.                  */
/*                                                                           */
/*****************************************************************************/

static KHE_FRAMED_MONITOR KheFramedMonitorMake(KHE_MONITOR m,
  int first_index, int last_index, HA_ARENA a)
{
  KHE_FRAMED_MONITOR res;
  HaMake(res, a);
  res->monitor = m;
  res->first_index = first_index;
  res->last_index = last_index;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheFramedMonitorFree(KHE_FRAMED_MONITOR fm)                         */
/*                                                                           */
/*  Free fm.                                                                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheFramedMonitorFree(KHE_FRAMED_MONITOR fm)
{
  MFree(fm);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheFramedMonitorCmp(const void *t1, const void *t2)                  */
/*                                                                           */
/*  Untyped comparison function for sorting an array of framed monitor       */
/*  objects by increasing (first_index, last_index).                         */
/*                                                                           */
/*****************************************************************************/

int KheFramedMonitorCmp(const void *t1, const void *t2)
{
  KHE_FRAMED_MONITOR fm1 = * (KHE_FRAMED_MONITOR *) t1;
  KHE_FRAMED_MONITOR fm2 = * (KHE_FRAMED_MONITOR *) t2;
  int cmp;
  cmp = fm1->first_index - fm2->first_index;
  if( cmp != 0 ) return cmp;
  return fm1->last_index - fm2->last_index;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheUseClusterBusyTimesMonitor(KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm,  */
/*    KHE_FRAME common_frame, KHE_FRAMED_MONITOR *res)                       */
/*                                                                           */
/*  If cbtm is relevant, return true and set *res to its framed monitor      */
/*  object.  Otherwise return false.                                         */
/*                                                                           */
/*****************************************************************************/

static bool KheUseClusterBusyTimesMonitor(KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm,
  KHE_FRAME common_frame, KHE_FRAMED_MONITOR *res, HA_ARENA a)
{
  int i;  KHE_TIME_GROUP tg;  KHE_POLARITY po;  KHE_RESOURCE_GROUP rg;
  KHE_TIME t, earliest_t, latest_t;  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT cbtc;

  /* if monitor's constraint does not monitor all resources, return false */
  cbtc = KheClusterBusyTimesMonitorConstraint(cbtm);
  if( KheClusterBusyTimesConstraintResourceGroupCount(cbtc) != 1 )
    return *res = NULL, false;
  rg = KheClusterBusyTimesConstraintResourceGroup(cbtc, 0);
  if( KheResourceGroupResourceCount(rg) !=
      KheResourceTypeResourceCount(KheResourceGroupResourceType(rg)) )
    return *res = NULL, false;

  /* find the earliest and latest times; if none, return false */
  earliest_t = latest_t = NULL;
  for( i = 0;  i < KheClusterBusyTimesMonitorTimeGroupCount(cbtm);  i++ )
  {
    tg = KheClusterBusyTimesMonitorTimeGroup(cbtm, i, &po);
    if( KheTimeGroupTimeCount(tg) > 0 )
    {
      t = KheTimeGroupTime(tg, 0);
      if( earliest_t == NULL || KheTimeIndex(t) < KheTimeIndex(earliest_t) )
        earliest_t = t;
      t = KheTimeGroupTime(tg, KheTimeGroupTimeCount(tg) - 1);
      if( latest_t == NULL || KheTimeIndex(t) > KheTimeIndex(latest_t) )
        latest_t = t;
    }
  }
  if( earliest_t == NULL )
    return *res = NULL, false;

  /* all good, return a framed monitor object */
  return *res = KheFramedMonitorMake((KHE_MONITOR) cbtm,
    KheFrameTimeIndex(common_frame, earliest_t),
    KheFrameTimeIndex(common_frame, latest_t), a), true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheUseLimitBusyTimesMonitor(KHE_LIMIT_BUSY_TIMES_MONITOR lbtm,      */
/*    KHE_FRAME common_frame, KHE_FRAMED_MONITOR *res)                       */
/*                                                                           */
/*  If lbtm is relevant, return true and set *res to its framed monitor      */
/*  object.  Otherwise return false.                                         */
/*                                                                           */
/*****************************************************************************/

static bool KheUseLimitBusyTimesMonitor(KHE_LIMIT_BUSY_TIMES_MONITOR lbtm,
  KHE_FRAME common_frame, KHE_FRAMED_MONITOR *res, HA_ARENA a)
{
  int i, offset;  KHE_TIME_GROUP tg;  KHE_RESOURCE_GROUP rg;
  KHE_TIME t, earliest_t, latest_t;  KHE_LIMIT_BUSY_TIMES_CONSTRAINT lbtc;

  /* if monitor's constraint does not monitor all resources, return false */
  lbtc = KheLimitBusyTimesMonitorConstraint(lbtm);
  if( KheLimitBusyTimesConstraintResourceGroupCount(lbtc) != 1 )
    return *res = NULL, false;
  rg = KheLimitBusyTimesConstraintResourceGroup(lbtc, 0);
  if( KheResourceGroupResourceCount(rg) !=
      KheResourceTypeResourceCount(KheResourceGroupResourceType(rg)) )
    return *res = NULL, false;

  /* find the earliest and latest times; if none, return false */
  earliest_t = latest_t = NULL;
  offset = KheLimitBusyTimesMonitorOffset(lbtm);
  for( i = 0;  i < KheLimitBusyTimesConstraintTimeGroupCount(lbtc);  i++ )
  {
    tg = KheLimitBusyTimesConstraintTimeGroup(lbtc, i, offset);
    if( KheTimeGroupTimeCount(tg) > 0 )
    {
      t = KheTimeGroupTime(tg, 0);
      if( earliest_t == NULL || KheTimeIndex(t) < KheTimeIndex(earliest_t) )
        earliest_t = t;
      t = KheTimeGroupTime(tg, KheTimeGroupTimeCount(tg) - 1);
      if( latest_t == NULL || KheTimeIndex(t) > KheTimeIndex(latest_t) )
        latest_t = t;
    }
  }
  if( earliest_t == NULL )
    return *res = NULL, false;

  /* all good, return a framed monitor object */
  return *res = KheFramedMonitorMake((KHE_MONITOR) lbtm,
    KheFrameTimeIndex(common_frame, earliest_t),
    KheFrameTimeIndex(common_frame, latest_t), a), true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheUseMonitor(KHE_MONITOR m, KHE_FRAME common_frame,                */
/*    KHE_FRAMED_MONITOR *res)                                               */
/*                                                                           */
/*  If m is relevant, return true and set *res to its framed monitor         */
/*  object.  Otherwise return false.                                         */
/*                                                                           */
/*****************************************************************************/

static bool KheUseMonitor(KHE_MONITOR m, KHE_FRAME common_frame,
  KHE_FRAMED_MONITOR *res, HA_ARENA a)
{
  switch( KheMonitorTag(m) )
  {
    case KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG:

      return KheUseClusterBusyTimesMonitor((KHE_CLUSTER_BUSY_TIMES_MONITOR) m,
        common_frame, res, a);

    case KHE_LIMIT_BUSY_TIMES_MONITOR_TAG:

      return KheUseLimitBusyTimesMonitor((KHE_LIMIT_BUSY_TIMES_MONITOR) m,
        common_frame, res, a);

    default:

      /* ignore other kinds of monitors */
      return *res = NULL, false;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheFramedMonitorDebug(KHE_FRAMED_MONITOR fm, int indent, FILE *fp)  */
/*                                                                           */
/*  Debug print of fm onto fp with the given indent.                         */
/*                                                                           */
/*****************************************************************************/

static void KheFramedMonitorDebug(KHE_FRAMED_MONITOR fm, int indent, FILE *fp)
{
  char buff[50];
  sprintf(buff, " %d-%d: ", fm->first_index, fm->last_index);
  fprintf(fp, "%*s%-12s", indent, "", buff);
  KheMonitorDebug(fm->monitor, 2, 0, fp);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "time sets"                                                    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_TIME_SET KheTimeSetMake(void)                                        */
/*                                                                           */
/*  Make a new, empty time set object.                                       */
/*                                                                           */
/*****************************************************************************/

static KHE_TIME_SET KheTimeSetMake(HA_ARENA a)
{
  KHE_TIME_SET res;
  HaMake(res, a);
  HaArrayInit(res->times, a);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheTimeSetTimeCount(KHE_TIME_SET ts)                                 */
/*                                                                           */
/*  Return the number of times in ts.                                        */
/*                                                                           */
/*****************************************************************************/

static int KheTimeSetTimeCount(KHE_TIME_SET ts)
{
  return HaArrayCount(ts->times);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheTimeSetClear(KHE_TIME_SET ts)                                    */
/*                                                                           */
/*  Clear (but do not delete) ts.                                            */
/*                                                                           */
/*****************************************************************************/

static void KheTimeSetClear(KHE_TIME_SET ts)
{
  HaArrayClear(ts->times);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheTimeSetCopy(KHE_TIME_SET target, KHE_TIME_SET source)            */
/*                                                                           */
/*  Clear out whatever was in target, and copy source into it.               */
/*                                                                           */
/*****************************************************************************/

static void KheTimeSetCopy(KHE_TIME_SET target, KHE_TIME_SET source)
{
  int i;
  HaArrayClear(target->times);
  HaArrayAppend(target->times, source->times, i);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheTimeSetFree(KHE_TIME_SET ts)                                     */
/*                                                                           */
/*  Free ts.                                                                 */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheTimeSetFree(KHE_TIME_SET ts)
{
  MArrayFree(ts->times);
  MFree(ts);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheTimeSetDebug(KHE_TIME_SET ts, int indent, FILE *fp)              */
/*                                                                           */
/*  Debug print of time set ts.  This function accepts NULL for ts.          */
/*                                                                           */
/*****************************************************************************/

static void KheTimeSetDebug(KHE_TIME_SET ts, int indent, FILE *fp)
{
  KHE_TIME t;  int i;
  fprintf(fp, "%*s", indent, "");
  if( ts == NULL )
    fprintf(fp, "-");
  else
  {
    fprintf(fp, "{");
    HaArrayForEach(ts->times, t, i)
      fprintf(fp, "%s%s", i > 0 ? ", " : "", KheTimeId(t));
    fprintf(fp, "}");
  }
  fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "group by frame solvers"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_GROUP_BY_FRAME_SOLVER KheGroupByFrameSolverMake(                     */
/*    KHE_TASKING tasking, int max_days, bool resource_invariant,            */
/*    KHE_FRAME common_frame, KHE_EVENT_TIMETABLE_MONITOR etm)               */
/*                                                                           */
/*  Make a new group by frame solver object.                                 */
/*                                                                           */
/*****************************************************************************/

static KHE_GROUP_BY_FRAME_SOLVER KheGroupByFrameSolverMake(
  KHE_TASKING tasking, int max_days, bool resource_invariant,
  KHE_FRAME common_frame, KHE_EVENT_TIMETABLE_MONITOR etm,
  KHE_TASK_SET ts, HA_ARENA a)
{
  KHE_GROUP_BY_FRAME_SOLVER res;  /* HA_ARENA a; */

  /* the solve as a whole */
  /* a = HaArenaM ake(); */
  HaMake(res, a);
  res->arena = a;
  res->tasking = tasking;
  res->soln = KheTaskingSoln(tasking);
  res->resource_invariant = resource_invariant;
  res->common_frame = common_frame;
  res->etm = etm;
  res->resource = NULL;
  res->soln_changed = false;
  res->task_set = ts;

  /* phase 1 */
  res->max_days = max_days;
  HaArrayInit(res->framed_monitors, a);
  HaArrayInit(res->start_indexes, a);
  res->group_monitor = NULL;
  res->last_index = 0;
  res->success_count = 0;
  res->success_ts = KheTimeSetMake(a);
  res->curr_ts = KheTimeSetMake(a);

  /* phase 2 */
  HaArrayInit(res->time_sets, a);
  HaArrayInit(res->time_sets_by_time, a);
  HaArrayInit(res->time_sets_by_monitor, a);
  HaArrayInit(res->success_seq, a);
  HaArrayInit(res->curr_seq, a);
  /* HaArrayInit(res->limit_active_intervals_monitors); */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFrameSolverFree(KHE_GROUP_BY_FRAME_SOLVER gfs)            */
/*                                                                           */
/*  Free gfs.                                                                */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheGroupByFrameSolverFree(KHE_GROUP_BY_FRAME_SOLVER gfs)
{
  HaArenaDel ete(gfs->arena);
  ** ***
  KHE_FRAMED_MONITOR fm;  int i;  KHE_TIME_SET ts;

  ** phase 1 **
  HaArrayForEach(gfs->framed_monitors, fm, i)
    KheFramedMonitorFree(fm);
  MArrayFree(gfs->framed_monitors);
  MArrayFree(gfs->start_indexes);
  HnAssert(gfs->group_monitor == NULL,
    "KheGroupByFrameSolverFree internal error");
  KheTimeSetFree(gfs->success_ts);
  KheTimeSetFree(gfs->curr_ts);

  ** phase 2 **
  HaArrayForEach(gfs->time_sets, ts, i)
    KheTimeSetFree(ts);
  MArrayFree(gfs->time_sets);
  MArrayFree(gfs->time_sets_by_time);
  MArrayFree(gfs->time_sets_by_monitor);
  MArrayFree(gfs->success_seq);
  MArrayFree(gfs->curr_seq);
  ** MArrayFree(gfs->limit_active_intervals_monitors); **
  MFree(gfs);
  *** **
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  int KheTaskCmp(const void *t1, const void *t2)                           */
/*                                                                           */
/*  Comparison function for sorting tasks so that they can be uniqueified.   */
/*                                                                           */
/*****************************************************************************/

static int KheTaskCmp(const void *t1, const void *t2)
{
  KHE_TASK task1 = * (KHE_TASK *) t1;
  KHE_TASK task2 = * (KHE_TASK *) t2;
  return KheTaskSolnIndex(task1) - KheTaskSolnIndex(task2);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_TASK_SET KheFindAllTasksAtTime(KHE_GROUP_BY_FRAME_SOLVER gfs,        */
/*    KHE_TIME time)                                                         */
/*                                                                           */
/*  Return a new task set containing all the tasks of duration 1 running     */
/*  at time, or rather, their proper roots, uniqueified.                     */
/*                                                                           */
/*****************************************************************************/

static KHE_TASK_SET KheFindAllTasksAtTime(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_TIME time)
{
  KHE_TASK_SET res;  KHE_MEET meet;  KHE_TASK task;  int i, j;
  KHE_RESOURCE_TYPE rt;
  rt = KheResourceResourceType(gfs->resource);
  res = KheTaskSetMake(gfs->soln);
  for( i = 0; i < KheEventTimetableMonitorTimeMeetCount(gfs->etm, time); i++ )
  {
    meet = KheEventTimetableMonitorTimeMeet(gfs->etm, time, i);
    if( KheMeetDuration(meet) == 1 )
      for( j = 0;  j < KheMeetTaskCount(meet);  j++ )
      {
	task = KheTaskProperRoot(KheMeetTask(meet, j));
	if( KheTaskResourceType(task) == rt )
	  KheTaskSetAddTask(res, task);
      }
  }
  KheTaskSetSortUnique(res, &KheTaskCmp);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheFindAndAssignTaskAtTime(KHE_GROUP_BY_FRAME_SOLVER gfs,           */
/*    KHE_TIME time, KHE_TASK *task)                                         */
/*                                                                           */
/*  Find and assign an arbitrary unassigned task of duration 1 at this       */
/*  time, and return true if successful.                                     */
/*                                                                           */
/*****************************************************************************/

static bool KheFindAndAssignTaskAtTime(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_TIME time, KHE_TASK *task)
{
  KHE_MEET meet;  KHE_TASK task2;  int i, j;
  for( i = 0; i < KheEventTimetableMonitorTimeMeetCount(gfs->etm, time); i++ )
  {
    meet = KheEventTimetableMonitorTimeMeet(gfs->etm, time, i);
    if( KheMeetDuration(meet) == 1 )
      for( j = 0;  j < KheMeetTaskCount(meet);  j++ )
      {
	task2 = KheTaskProperRoot(KheMeetTask(meet, j));
	if( KheTaskAssignResource(task2, gfs->resource) )
	  return *task = task2, true;
      }
  }
  return *task = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "phase 1"                                                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFrameAssign(KHE_GROUP_BY_FRAME_SOLVER gfs, int curr_index)*/
/*                                                                           */
/*  Carry out the part of the solve that assigns all combinations of tasks   */
/*  and checks which ones are successful.                                    */
/*                                                                           */
/*****************************************************************************/

static void KheGroupByFrameAssign(KHE_GROUP_BY_FRAME_SOLVER gfs, int curr_index)
{
  KHE_TIME_GROUP tg;  int i;  KHE_POLARITY po;  KHE_TIME t;  KHE_TASK task;
  if( curr_index > gfs->last_index )
  {
    /* curr_ts is complete; record it if successful (has zero cost) */
    if( KheMonitorCost((KHE_MONITOR) gfs->group_monitor) == 0 )
    {
      /* curr_ts is a success */
      gfs->success_count++;
      KheTimeSetCopy(gfs->success_ts, gfs->curr_ts);
    }
  }
  else
  {
    /* try assigning at time in the current time group */
    tg = KheFrameTimeGroup(gfs->common_frame, curr_index, &po);
    for( i = 0;  gfs->success_count < 2 && i < KheTimeGroupTimeCount(tg);  i++ )
    {
      t = KheTimeGroupTime(tg, i);
      if( KheFindAndAssignTaskAtTime(gfs, t, &task) )
      {
	HaArrayAddLast(gfs->curr_ts->times, t);
        KheGroupByFrameAssign(gfs, curr_index + 1);
	HaArrayDeleteLast(gfs->curr_ts->times);
	KheTaskUnAssign(task);
      }
    }

    /* try not assigning within the current time group */
    KheGroupByFrameAssign(gfs, curr_index + 1);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFrameGroupTasks(KHE_GROUP_BY_FRAME_SOLVER gfs,            */
/*     KHE_TIME_SET ts)                                                      */
/*                                                                           */
/*  Group the tasks identified by ts, and set gfs->soln_changed if some      */
/*  actual grouping occurs.                                                  */
/*                                                                           */
/*****************************************************************************/
typedef HA_ARRAY(KHE_TASK_SET) ARRAY_KHE_TASK_SET;

static void KheGroupByFrameGroupTasks(KHE_GROUP_BY_FRAME_SOLVER gfs,
   KHE_TIME_SET ts)
{
  ARRAY_KHE_TASK_SET task_sets;  KHE_TASK_SET task_set;  KHE_TIME t;
  KHE_TASK first_task, task;  int i, j, count, min_count;  bool ok;
  if( KheTimeSetTimeCount(ts) >= 2 )
  {
    /* get one task set for each time; also calculate the smallest size */
    HaArrayInit(task_sets, gfs->arena);
    min_count = INT_MAX;
    HaArrayForEach(ts->times, t, i)
    {
      task_set = KheFindAllTasksAtTime(gfs, t);
      HaArrayAddLast(task_sets, task_set);
      count = KheTaskSetTaskCount(task_set);
      if( count < min_count )
	min_count = count;
    }

    /* carry out the groupings; NB none of the tasks are currently assigned */
    for( i = 0;  i < min_count;  i++ )
    {
      /* group together the i'th tasks from each task set */
      task_set = HaArray(task_sets, 0);
      first_task = KheTaskSetTask(task_set, i);
      if( DEBUG3 )
      {
	fprintf(stderr, "    KheGroupByFrame grouping tasks {");
	KheTaskDebug(first_task, 1, -1, stderr);
      }
      for( j = 1;  j < HaArrayCount(task_sets);  j++ )
      {
	task_set = HaArray(task_sets, j);
	task = KheTaskSetTask(task_set, i);
	if( task == first_task )
	  ok = true;
	else
	{
	  ok = KheTaskAssign(task, first_task);
	  if( ok )
	  {
	    gfs->soln_changed = true;
	    if( gfs->task_set != NULL )
	      KheTaskSetAddTask(gfs->task_set, task);
	  }
	}
	if( DEBUG3 )
	{
	  fprintf(stderr, ", %s", ok ? "" : "not ");
	  KheTaskDebug(task, 1, -1, stderr);
	}
      }
      if( DEBUG3 )
	fprintf(stderr, "}\n");
    }

    /* free the task sets and the array of task sets */
    HaArrayForEach(task_sets, task_set, i)
      KheTaskSetDelete(task_set);
    HaArrayFree(task_sets);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFramePhase1ForInterval(KHE_GROUP_BY_FRAME_SOLVER gfs,     */
/*    int first_index, int last_index)                                       */
/*                                                                           */
/*  Solve gfs for the interval first_index .. last_index, and return         */
/*  true if there was exactly one success, with gfs->success_ts holding      */
/*  the successful set of times.                                             */
/*                                                                           */
/*****************************************************************************/

static void KheGroupByFramePhase1ForInterval(KHE_GROUP_BY_FRAME_SOLVER gfs,
  int first_index, int last_index)
{
  int i, si;  KHE_FRAMED_MONITOR fm;  KHE_TIME_GROUP tg;  KHE_POLARITY po;
  KHE_TIME t;  KHE_TASK task;  KHE_TIME_SET ts;
  if( DEBUG2 )
    fprintf(stderr,
      "  [ KheGroupByFrameSolverFindAsst(gfs, first %d, last %d)\n",
      first_index, last_index);

  /* gather relevant monitors into a group monitor */
  gfs->group_monitor = KheGroupMonitorMake(gfs->soln, 1, "GroupByFrame");
  si = (first_index < HaArrayCount(gfs->start_indexes) ?
    HaArray(gfs->start_indexes, first_index) :
    HaArrayCount(gfs->framed_monitors));
  for( i = si;  i < HaArrayCount(gfs->framed_monitors);  i++ )
  {
    /* fm starts at or after first_index */
    fm = HaArray(gfs->framed_monitors, i);
    HnAssert(fm->first_index >= first_index,
      "KheGroupByFrameSolverFindAsst internal error");
    if( fm->last_index <= last_index )
    {
      KheGroupMonitorAddChildMonitor(gfs->group_monitor, fm->monitor);
      if( DEBUG2 )
      {
	fprintf(stderr, "    adding ");
	KheFramedMonitorDebug(fm, 0, stderr);
      }
    }

    /* if fm->first_index > last_index, there is no point in going further */
    if( fm->first_index > last_index )
      break;
  }

  /* try one solve for each time in the first time group */
  tg = KheFrameTimeGroup(gfs->common_frame, first_index, &po);
  for( i = 0;  i < KheTimeGroupTimeCount(tg);  i++ )
  {
    /* initialize the solve fields other than gm */
    gfs->last_index = last_index;
    gfs->success_count = 0;
    KheTimeSetClear(gfs->success_ts);
    KheTimeSetClear(gfs->curr_ts);

    /* solve for t */
    t = KheTimeGroupTime(tg, i);
    if( KheFindAndAssignTaskAtTime(gfs, t, &task) )
    {
      HaArrayAddLast(gfs->curr_ts->times, t);
      KheGroupByFrameAssign(gfs, first_index + 1);
      HaArrayDeleteLast(gfs->curr_ts->times);
      KheTaskUnAssign(task);
    }

    /* if exactly one success, we have succeeded */
    if( gfs->success_count == 1 )
    {
      /* have a single successful assignment, so group tasks */
      if( DEBUG2 )
      {
	fprintf(stderr, "  KheGroupByFrameSolverFindAsst found group: ");
	KheTimeSetDebug(gfs->success_ts, 0, stderr);
      }
      ts = KheTimeSetMake(gfs->arena);
      KheTimeSetCopy(ts, gfs->success_ts);
      HaArrayAddLast(gfs->time_sets, ts);
      KheGroupByFrameGroupTasks(gfs, ts);
    }
  }

  /* clear out for next (rt, f), and return */
  KheGroupMonitorDelete(gfs->group_monitor);
  gfs->group_monitor = NULL;
  if( DEBUG2 )
    fprintf(stderr, "  ] KheGroupByFrameSolverFindAsst returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFrameSolverDebug(KHE_GROUP_BY_FRAME_SOLVER gfs,           */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Produce a debug print of gfs on fp with the given indent.                */
/*                                                                           */
/*****************************************************************************/

static void KheGroupByFrameSolverDebug(KHE_GROUP_BY_FRAME_SOLVER gfs,
  int indent, FILE *fp)
{
  KHE_FRAMED_MONITOR fm;  int i, pos;
  fprintf(fp, "%*s[ GroupByFrameSolver(%s, r %s, max_days %d, rs_invt %s)\n", 
    indent, "", KheInstanceId(KheSolnInstance(gfs->soln)),
    KheResourceId(gfs->resource), gfs->max_days,
    gfs->resource_invariant ? "true" : "false");
  HaArrayForEach(gfs->framed_monitors, fm, i)
  {
    fprintf(fp, "%*s%2d: ", indent, "", i);
    KheFramedMonitorDebug(fm, 0, fp);
  }
  fprintf(fp, "%*s[", indent + 2, "");
  HaArrayForEach(gfs->start_indexes, pos, i)
    fprintf(fp, "%s%d", i > 0 ? ", " : "", pos);
  fprintf(fp, "]\n");
  fprintf(fp, "%*s]\n", indent, "");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFramePhase1(KHE_GROUP_BY_FRAME_SOLVER gfs)                */
/*                                                                           */
/*  Carry out group by frame phase 1.  Field gfs->resource is non-NULL       */
/*  and determines which resource type, and which monitors, to use.          */
/*                                                                           */
/*****************************************************************************/

static void KheGroupByFramePhase1(KHE_GROUP_BY_FRAME_SOLVER gfs)
{
  KHE_FRAMED_MONITOR fm;  int i, pos, si, len, first_index, days;
  KHE_MONITOR m;

  /* build up and sort the framed_monitors array */
  HnAssert(HaArrayCount(gfs->framed_monitors) == 0,
    "KheGroupByFramePhase1 internal error 1");
  for( i = 0;  i < KheSolnResourceMonitorCount(gfs->soln, gfs->resource);  i++ )
  {
    m = KheSolnResourceMonitor(gfs->soln, gfs->resource, i);
    if( KheUseMonitor(m, gfs->common_frame, &fm, gfs->arena) )
      HaArrayAddLast(gfs->framed_monitors, fm);
  }
  HaArraySort(gfs->framed_monitors, &KheFramedMonitorCmp);

  /* build up the start_indexes array */
  HnAssert(HaArrayCount(gfs->start_indexes) == 0,
    "KheGroupByFramePhase1 internal error 1");
  pos = 0;
  for( si = 0;  pos < HaArrayCount(gfs->framed_monitors);  si++ )
  {
    /* add the index for si */
    fm = HaArray(gfs->framed_monitors, pos);
    while( fm->first_index < si )
    {
      pos++;
      if( pos >= HaArrayCount(gfs->framed_monitors) )
	break;
      fm = HaArray(gfs->framed_monitors, pos);
    }
    HaArrayAddLast(gfs->start_indexes, pos);
  }

  if( DEBUG2 )
    KheGroupByFrameSolverDebug(gfs, 2, stderr);

  /* solve for each subsequence of up to max_days days */
  days = KheFrameTimeGroupCount(gfs->common_frame);
  for( len = gfs->max_days;  len >= 2;  len-- )
    for( first_index = 0;  first_index <= days - len;  first_index++ )
      KheGroupByFramePhase1ForInterval(gfs, first_index, first_index + len-1);

  /* clear out the framed_monitors and start_indexes arrays */
  /* ***
  HaArrayForEach(gfs->framed_monitors, fm, i)
    KheFramedMonitorFree(fm);
  *** */
  HaArrayClear(gfs->framed_monitors);
  HaArrayClear(gfs->start_indexes);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "phase 2"                                                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheLimitActiveIntervalsMonitorIsSuitable(                           */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim)                               */
/*                                                                           */
/*  Return true if laim is suitable for Phase 2:  if its constraint applies  */
/*  to all resources, and all its time groups are singletons and positive.   */
/*                                                                           */
/*****************************************************************************/

static bool KheLimitActiveIntervalsMonitorIsSuitable(
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim)
{
  int i;  KHE_TIME_GROUP tg;  KHE_POLARITY po;  KHE_RESOURCE_GROUP rg;
  KHE_LIMIT_ACTIVE_INTERVALS_CONSTRAINT laic;

  /* if monitor's constraint does not monitor all resources, return false */
  laic = KheLimitActiveIntervalsMonitorConstraint(laim);
  if( KheLimitActiveIntervalsConstraintResourceGroupCount(laic) != 1 )
    return false;
  rg = KheLimitActiveIntervalsConstraintResourceGroup(laic, 0);
  if( KheResourceGroupResourceCount(rg) !=
      KheResourceTypeResourceCount(KheResourceGroupResourceType(rg)) )
    return false;

  /* if one of its time groups is not single or not positive, return false */
  for( i = 0;  i < KheLimitActiveIntervalsMonitorTimeGroupCount(laim);  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, i, &po);
    if( KheTimeGroupTimeCount(tg) != 1 || po != KHE_POSITIVE )
      return false;
  }

  /* all OK */
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  void FindSeq(KHE_GROUP_BY_FRAME_SOLVER gfs, int min_limit,               */
/*    int max_limit, int remaining_len, int *node_count)                     */
/*                                                                           */
/*  This is the recursive part of finding a unique sequence of               */
/*  sub-interval lengths, if there is one.  Each length must lie             */
/*  between min_limit and max_limit; the remaining wanted length             */
/*  is remaining_len.                                                        */
/*                                                                           */
/*****************************************************************************/

static void FindSeq(KHE_GROUP_BY_FRAME_SOLVER gfs, int min_limit,
  int max_limit, int remaining_len, int *node_count)
{
  int first_len, len, i;
  if( ++(*node_count) > 100 )
  {
    /* give up, search is taking too long */
    gfs->success_count = 2;
  }
  else if( remaining_len < 0 )
  {
    /* no solutions here */
  }
  else if( remaining_len == 0 )
  {
    /* curr_seq is a feasible soln; but no good if more than one */
    gfs->success_count++;
    if( gfs->success_count >= 2 )
      return;

    /* if any of the lengths of curr_seq differ, there is more than one soln */
    first_len = HaArrayFirst(gfs->curr_seq);
    for( i = 1;  i < HaArrayCount(gfs->curr_seq);  i++ )
      if( HaArray(gfs->curr_seq, i) != first_len )
      {
	gfs->success_count = 2;
	return;
      }

    /* save curr_seq; at the moment it is the unique feasible solution */
    HaArrayClear(gfs->success_seq);
    HaArrayForEach(gfs->curr_seq, len, i)
      HaArrayAddLast(gfs->success_seq, len);
  }
  else /* remaining_len > 0 */
  {
    for( len = min_limit;  len <= max_limit && gfs->success_count < 2;  len++ )
    {
      HaArrayAddLast(gfs->curr_seq, len);
      FindSeq(gfs, min_limit, max_limit, remaining_len - len, node_count);
      HaArrayDeleteLast(gfs->curr_seq);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMergeOneGap(KHE_GROUP_BY_FRAME_SOLVER gfs,                       */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, int gap_start, int gap_stop)  */
/*                                                                           */
/*  Do merges in the gap from gap_start inclusive to gap_stop exclusive.     */
/*                                                                           */
/*****************************************************************************/

static void KheMergeOneGap(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim, int gap_start, int gap_stop)
{
  int i, j, len, min_limit, max_limit, node_count, start;
  KHE_TIME_GROUP tg;  KHE_POLARITY po;  KHE_TIME_SET ts;

  min_limit = KheLimitActiveIntervalsMonitorMinimum(laim);
  max_limit = KheLimitActiveIntervalsMonitorMaximum(laim);
  if( DEBUG4 )
  {
    fprintf(stderr, "  [ KheMergeOneGap(gfs, laim(%d .. %d), %d, %d):\n",
      min_limit, max_limit, gap_start, gap_stop);
    for( i = gap_start;  i < gap_stop;  i++ )
    {
      tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, i, &po);
      KheTimeGroupDebug(tg, 1, 6, stderr);
    }
  }

  gfs->success_count = 0;
  HaArrayClear(gfs->success_seq);
  HaArrayClear(gfs->curr_seq);
  FindSeq(gfs, min_limit, max_limit, gap_stop - gap_start, &node_count);
  if( DEBUG4 )
  {
    fprintf(stderr, "      found %d solution%s", gfs->success_count,
      gfs->success_count != 1 ? "s" : "");
    if( gfs->success_count == 1 )
    {
      fprintf(stderr, ":");
      HaArrayForEach(gfs->success_seq, len, i)
	fprintf(stderr, " %d", len);
    }
    fprintf(stderr, "\n");
  }
  if( gfs->success_count == 1 )
  {
    /* merge the sets defined by the successful sequence */
    start = gap_start;
    ts = KheTimeSetMake(gfs->arena);
    HaArrayForEach(gfs->success_seq, len, i)
    {
      KheTimeSetClear(ts);
      for( j = 0;  j < len;  j++ )
      {
	tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, start + j, &po);
	HaArrayAddLast(ts->times, KheTimeGroupTime(tg, 0));
      }
      if( DEBUG4 )
      {
	fprintf(stderr, "      merging time set ");
	KheTimeSetDebug(ts, 0, stderr);
      }
      /* actual merge still to do */
      KheGroupByFrameGroupTasks(gfs, ts);
      start += len;
    }
    /* KheTimeSetFree(ts); */
  }

  if( DEBUG4 )
    fprintf(stderr, "  ] KheMergeOneGap returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupOneMonitor(KHE_GROUP_BY_FRAME_SOLVER gfs,                   */
/*    KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim)                               */
/*                                                                           */
/*  Carry out Phase 2 grouping based on laim.                                */
/*                                                                           */
/*****************************************************************************/

static void KheGroupOneMonitor(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim)
{
  int i, j, min_limit, tg_count, index;
  KHE_TIME_GROUP tg;  KHE_POLARITY po;  KHE_TIME t;  KHE_TIME_SET ts;
  if( DEBUG4 )
  {
    fprintf(stderr, "[ KheGroupOneMonitor(gfs, laim)\n");
    KheMonitorDebug((KHE_MONITOR) laim, 1, 2, stderr);
  }

  /* there must be at least two time groups */
  if( KheLimitActiveIntervalsMonitorTimeGroupCount(laim) < 2 )
    return;

  /* set up the time_sets_by_monitor array */
  if( DEBUG4 )
    fprintf(stderr, "    [ time_sets_by_monitor:\n");
  HaArrayClear(gfs->time_sets_by_monitor);
  tg_count = KheLimitActiveIntervalsMonitorTimeGroupCount(laim);
  for( i = 0;  i < tg_count;  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, i, &po);
    HnAssert(KheTimeGroupTimeCount(tg) == 1 && po == KHE_POSITIVE,
      "KheGroupOneMonitor internal error");
    t = KheTimeGroupTime(tg, 0);
    ts = HaArray(gfs->time_sets_by_time, KheTimeIndex(t));
    HaArrayAddLast(gfs->time_sets_by_monitor, ts);
    if( DEBUG4 )
    {
      fprintf(stderr, "        %d: ", i);
      KheTimeGroupDebug(tg, 1, -1, stderr);
      fprintf(stderr, " --> ");
      KheTimeSetDebug(ts, 0, stderr);
    }
  }
  if( DEBUG4 )
    fprintf(stderr, "    ]\n");

  /* if time sets do not coordinate with time groups, exit early */
  for( i = 0;  i < tg_count;  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, i, &po);
    ts = HaArray(gfs->time_sets_by_monitor, i);
    if( ts != NULL )
    {
      for( j = i + 1;  j < tg_count;  j++ )
      {
	if( HaArray(gfs->time_sets_by_monitor, j) != ts )
	  break;
      }
      if( j - i != KheTimeSetTimeCount(ts) )
      {
	if( DEBUG4 )
	{
	  fprintf(stderr, "] KheGroupOneMonitor returning: uncoordinate ");
	  KheTimeSetDebug(ts, 0, stderr);
	}
	return;
      }
      i = j - 1;
    }
  }

  /* extend the grouped sequence at left, if there is one and it's too short */
  min_limit = KheLimitActiveIntervalsMonitorMinimum(laim);
  ts = HaArrayFirst(gfs->time_sets_by_monitor);
  if( ts != NULL && KheLimitActiveIntervalsMonitorHistoryBefore(laim) <= 0 &&
      KheTimeSetTimeCount(ts) < min_limit )
  {
    for( i = KheTimeSetTimeCount(ts);  i < min_limit && i < tg_count;  i++ )
    {
      index = i;
      if( HaArray(gfs->time_sets_by_monitor, index) != NULL )
	break;  /* we could merge here, but we don't */
      tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, index, &po);
      t = KheTimeGroupTime(tg, 0);
      HaArrayPut(gfs->time_sets_by_monitor, index, ts);
      HaArrayAddLast(ts->times, t);
    }
    if( i > KheTimeSetTimeCount(ts) )
      KheGroupByFrameGroupTasks(gfs, ts);
  }

  /* extend the grouped sequence at right, if there is one and it's too short */
  ts = HaArrayLast(gfs->time_sets_by_monitor);
  if( ts != NULL && KheLimitActiveIntervalsMonitorHistoryAfter(laim) <= 0 &&
      KheTimeSetTimeCount(ts) < KheLimitActiveIntervalsMonitorMinimum(laim) )
  {
    for( i = KheTimeSetTimeCount(ts);  i < min_limit && i < tg_count;  i++ )
    {
      index = (tg_count - 1) - i;
      if( HaArray(gfs->time_sets_by_monitor, index) != NULL )
	break;  /* we could merge here, but we don't */
      tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, index, &po);
      t = KheTimeGroupTime(tg, 0);
      HaArrayPut(gfs->time_sets_by_monitor, index, ts);
      HaArrayAddLast(ts->times, t);
    }
    if( i > KheTimeSetTimeCount(ts) )
      KheGroupByFrameGroupTasks(gfs, ts);
  }

  /* find all gap intervals */
  for( i = 0;  i < KheLimitActiveIntervalsMonitorTimeGroupCount(laim);  i++ )
  {
    tg = KheLimitActiveIntervalsMonitorTimeGroup(laim, i, &po);
    ts = HaArray(gfs->time_sets_by_monitor, i);
    if( ts == NULL )
    {
      /* find gap i .. j - 1 */
      for( j=i+1; j < KheLimitActiveIntervalsMonitorTimeGroupCount(laim); j++ )
	if( HaArray(gfs->time_sets_by_monitor, j) != NULL )
	  break;

      /* merge within the gap, then carry on */
      KheMergeOneGap(gfs, laim, i, j);
      i = j - 1;
    }
  }


  if( DEBUG4 )
    fprintf(stderr, "] KheGroupOneMonitor returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFramePhase2(KHE_GROUP_BY_FRAME_SOLVER gfs)                */
/*                                                                           */
/*  Carry out Phase 2 of group by frame.                                     */
/*                                                                           */
/*****************************************************************************/

static void KheGroupByFramePhase2(KHE_GROUP_BY_FRAME_SOLVER gfs)
{
  KHE_MONITOR m;  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim;
  int i, j, time_count;  KHE_TIME_SET ts;  KHE_TIME t;  KHE_INSTANCE ins;

  /* sort and merge the time sets from Phase 1 */
  /* still to do - in GPost they are all disjoint anyway */
  /* when we do this, we can combine it with the next step */

  /* set up ts_indexes array, mapping times to time sets where present */
  HaArrayClear(gfs->time_sets_by_time);
  ins = KheSolnInstance(gfs->soln);
  time_count = KheInstanceTimeCount(ins);
  HaArrayFill(gfs->time_sets_by_time, time_count, NULL);
  HaArrayForEach(gfs->time_sets, ts, i)
    HaArrayForEach(ts->times, t, j)
      HaArrayPut(gfs->time_sets_by_time, KheTimeIndex(t), ts);
  if( DEBUG4 )
  {
    fprintf(stderr, "  [ time_sets_by_time:\n");
    HaArrayForEach(gfs->time_sets_by_time, ts, i)
    {
      fprintf(stderr, "    %d (%s): ", i, KheTimeId(KheInstanceTime(ins, i)));
      KheTimeSetDebug(ts, 0, stderr);
    }
    fprintf(stderr, "  ]\n");
  }

  /* do the job for each suitable limit active intervals monitor */
  for( i = 0;  i < KheSolnResourceMonitorCount(gfs->soln, gfs->resource);  i++ )
  {
    m = KheSolnResourceMonitor(gfs->soln, gfs->resource, i);
    if( KheMonitorTag(m) == KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG )
    {
      laim = (KHE_LIMIT_ACTIVE_INTERVALS_MONITOR) m;
      if( KheLimitActiveIntervalsMonitorIsSuitable(laim) )
        KheGroupOneMonitor(gfs, laim);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "main function"                                                */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  void KheGroupByFrameForResourceType(KHE_GROUP_BY_FRAME_SOLVER gfs,       */
/*    KHE_RESOURCE_TYPE rt)                                                  */
/*                                                                           */
/*  Carry out group by frame for resource type rt.  There is no point        */
/*  in doing anything if there are no resources, or only one resource.       */
/*                                                                           */
/*****************************************************************************/

static void KheGroupByFrameForResourceType(KHE_GROUP_BY_FRAME_SOLVER gfs,
  KHE_RESOURCE_TYPE rt)
{
  if( KheResourceTypeResourceCount(rt) > 1 )
  {
    gfs->resource = KheResourceTypeResource(rt, 0);
    KheGroupByFramePhase1(gfs);
    KheGroupByFramePhase2(gfs);
    gfs->resource = NULL;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheTaskingGroupByFrame(KHE_TASKING tasking, KHE_OPTIONS options,    */
/*    KHE_TASK_SET ts)                                                       */
/*                                                                           */
/*  Group the tasks of tasking by frame.                                     */
/*                                                                           */
/*  If ts is non-NULL, add every task that is given an assignment to ts.     */
/*                                                                           */
/*****************************************************************************/

bool KheTaskingGroupByFrame(KHE_TASKING tasking, KHE_OPTIONS options,
  KHE_TASK_SET ts)
{
  int max_days, i;  KHE_RESOURCE_TYPE rt;  bool res, resource_invariant;
  KHE_SOLN soln;  KHE_FRAME common_frame;  KHE_EVENT_TIMETABLE_MONITOR etm;
  KHE_INSTANCE ins;  KHE_GROUP_BY_FRAME_SOLVER gfs;  HA_ARENA a;

  /* return immediately if option rs_group_by_frame_off is true */
  soln = KheTaskingSoln(tasking);
  a = KheSolnArenaBegin(soln);
  ins = KheSolnInstance(soln);
  rt = KheTaskingResourceType(tasking);
  if( DEBUG1 )
    fprintf(stderr, "[ KheTaskingGroupByFrame(%s %s, -)\n", KheInstanceId(ins),
      rt == NULL ? "all resource types" : KheResourceTypeId(rt));
  if( KheOptionsGetBool(options, "rs_group_by_frame_off", false) )
  {
    if( DEBUG1 )
      fprintf(stderr, "] KheTaskingGroupByFrame returning false (off)\n");
    return false;
  }

  /* get the other options used by this function */
  max_days = KheOptionsGetInt(options, "rs_group_by_frame_max_days", 3);
  resource_invariant = KheOptionsGetBool(options, "rs_invariant", false);
  etm = (KHE_EVENT_TIMETABLE_MONITOR)
    KheOptionsGetObject(options, "gs_event_timetable_monitor", NULL);
  common_frame = KheOptionsFrame(options, "gs_common_frame", soln);
  HnAssert(etm != NULL, "KheTaskingGroupByFrame: required option "
    "gs_event_timetable_monitor not present");

  /* create a solver */
  gfs = KheGroupByFrameSolverMake(tasking, max_days, resource_invariant,
    common_frame, etm, ts, a);

  /* do group by frame for each resource type */
  if( rt != NULL )
    KheGroupByFrameForResourceType(gfs, rt);
  else
    for( i = 0;  i < KheInstanceResourceTypeCount(ins);  i++ )
      KheGroupByFrameForResourceType(gfs, KheInstanceResourceType(ins, i));

  /* free the solver and return */
  res = gfs->soln_changed;
  KheSolnArenaEnd(soln, a);
  /* KheGroupByFrameSolverFree(gfs); */
  if( DEBUG1 )
    fprintf(stderr, "] KheTaskingGroupByFrame returning %s\n",
      res ? "true" : "false");
  return res;
}
