
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sr_requested.c                                         */
/*  DESCRIPTION:  Satisfying requested task assignments                      */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"
#include <limits.h>

#define DEBUG1 0
#define DEBUG2 0
#define DEBUG3 0


typedef HA_ARRAY(KHE_LIMIT_BUSY_TIMES_MONITOR)
  ARRAY_KHE_LIMIT_BUSY_TIMES_MONITOR;

/*****************************************************************************/
/*                                                                           */
/*  int KheLimitBusyTimesMonitorDecreasingWeightCmp(const void *t1,          */
/*    const void *t2)                                                        */
/*                                                                           */
/*  Comparison function for sorting an array of limit busy times monitors    */
/*  by decreasing weight                                                     */
/*                                                                           */
/*****************************************************************************/

static int KheLimitBusyTimesMonitorDecreasingWeightCmp(const void *t1,
  const void *t2)
{
  KHE_COST weight1, weight2;
  KHE_LIMIT_BUSY_TIMES_CONSTRAINT lbtc1, lbtc2;
  KHE_LIMIT_BUSY_TIMES_MONITOR lbtm1 = * (KHE_LIMIT_BUSY_TIMES_MONITOR *) t1;
  KHE_LIMIT_BUSY_TIMES_MONITOR lbtm2 = * (KHE_LIMIT_BUSY_TIMES_MONITOR *) t2;
  lbtc1 = KheLimitBusyTimesMonitorConstraint(lbtm1);
  lbtc2 = KheLimitBusyTimesMonitorConstraint(lbtm2);
  weight1 = KheConstraintCombinedWeight((KHE_CONSTRAINT) lbtc1);
  weight2 = KheConstraintCombinedWeight((KHE_CONSTRAINT) lbtc2);
  return KheCostCmp(weight2, weight1);  /* reverse order, decreasing weight */
}


/*****************************************************************************/
/*                                                                           */
/*  void MakeBestAsst(KHE_SOLN soln, KHE_RESOURCE r,                         */
/*    KHE_RESOURCE_TIMETABLE_MONITOR rtm, KHE_TIME t,                        */
/*    KHE_EVENT_TIMETABLE_MONITOR etm, bool *soln_changed)                   */
/*                                                                           */
/*  Make the best possible assignment of a task to r (whose timetable        */
/*  monitor is rtm) at time t, that is, if r is not already busy then.       */
/*  Obtain the events running at time t from etm.  If an assignment is       */
/*  actually made, set *soln_changed to true.                                */
/*                                                                           */
/*  Actually, if the meet has more than 10 tasks, we only try every          */
/*  fifth task, cyclically starting from a random point.  This is because    */
/*  most tasks are symmetrical and it takes an unreasonably long time        */
/*  to try them all.                                                         */
/*                                                                           */
/*****************************************************************************/

static void MakeBestAsst(KHE_SOLN soln, KHE_RESOURCE r,
  KHE_RESOURCE_TIMETABLE_MONITOR rtm, KHE_TIME t,
  KHE_EVENT_TIMETABLE_MONITOR etm, bool *soln_changed)
{
  int i, j, count, extra, index, inc;  KHE_MEET meet;
  KHE_TASK task, best_task;  KHE_COST best_cost;

  /* if r is already busy at time t, return immediately */
  if( KheResourceTimetableMonitorTimeTaskCount(rtm, t) > 0 )
    return;

  /* find the best task among all events running at time t */
  count = 0;
  best_task = NULL;  best_cost = KheCost(INT_MAX, INT_MAX);
  extra = KheResourceInstanceIndex(r);
  for( i = 0;  i < KheEventTimetableMonitorTimeMeetCount(etm, t);  i++ )
  {
    meet = KheEventTimetableMonitorTimeMeet(etm, t, i);
    inc = (KheMeetTaskCount(meet) >= 10 ? 5 : 1);
    for( j = 0;  j < KheMeetTaskCount(meet);  j += inc )
    {
      index = (j + extra) % KheMeetTaskCount(meet);
      task = KheMeetTask(meet, index);
      count++;
      if( KheTaskAssignResource(task, r) )
      {
	if( KheSolnCost(soln) < best_cost )
	  best_task = task, best_cost = KheSolnCost(soln);
	KheTaskUnAssignResource(task);
      }
    }
  }

  /* if there is a best task, make that assignment and set *soln_changed */
  if( best_task != NULL )
  {
    KheTaskAssignResource(best_task, r);
    /* KheTaskAssignFix(best_task); doing this a better way now */
    *soln_changed = true;
  }
  if( DEBUG2 )
  {
    fprintf(stderr, "MakeBestAsst(soln, %s, %s) tried %d tasks: ",
      KheResourceId(r), KheTimeId(t), count);
    if( best_task != NULL )
      KheTaskDebug(best_task, 1, -1, stderr);
    else
      fprintf(stderr, "-");
    fprintf(stderr, "\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheSolnAssignRequestedResources(KHE_SOLN soln, KHE_OPTIONS options) */
/*                                                                           */
/*  Assign resources in soln as requested by its instance's limit busy       */
/*  times constraints.                                                       */
/*                                                                           */
/*  This function uses option gs_event_timetable_monitor.                    */
/*                                                                           */
/*****************************************************************************/

bool KheSolnAssignRequestedResources(KHE_SOLN soln, KHE_OPTIONS options)
{
  ARRAY_KHE_LIMIT_BUSY_TIMES_MONITOR lbt_monitors;  HA_ARENA a;
  KHE_INSTANCE ins;  KHE_RESOURCE r;  int i, j, k, offset, min_limit;
  KHE_MONITOR m;  KHE_LIMIT_BUSY_TIMES_MONITOR lbtm;  bool soln_changed;
  KHE_LIMIT_BUSY_TIMES_CONSTRAINT lbtc;  KHE_TIME_GROUP tg;  KHE_TIME t;
  KHE_RESOURCE_TIMETABLE_MONITOR rtm;  KHE_EVENT_TIMETABLE_MONITOR etm;

  /* do nothing if so instructed by the relevant option */
  if( KheOptionsGetBool(options, "rs_requested_off", false) )
    return false;

  /* get the event timetable monitor and abort if not available */
  if( DEBUG1 )
    fprintf(stderr, "[ KheSolnAssignRequestedResources(soln, options)\n");
  etm = (KHE_EVENT_TIMETABLE_MONITOR)
    KheOptionsGetObject(options, "gs_event_timetable_monitor", NULL);
  HnAssert(etm != NULL, "KheSolnAssignRequestedResources: "
    "no gs_event_timetable_monitor option in options");

  /* build and sort lbt_monitors, holding relevant limit busy times monitors */
  ins = KheSolnInstance(soln);
  a = KheSolnArenaBegin(soln);
  /* a = HaAren aMake(); */
  HaArrayInit(lbt_monitors, a);
  for( i = 0;  i < KheInstanceResourceCount(ins);  i++ )
  {
    r = KheInstanceResource(ins, i);
    for( j = 0;  j < KheSolnResourceMonitorCount(soln, r);  j++ )
    {
      m = KheSolnResourceMonitor(soln, r, j);
      if( KheMonitorTag(m) == KHE_LIMIT_BUSY_TIMES_MONITOR_TAG )
      {
	lbtm = (KHE_LIMIT_BUSY_TIMES_MONITOR) m;
	lbtc = KheLimitBusyTimesMonitorConstraint(lbtm);
	if( KheLimitBusyTimesConstraintMinimum(lbtc) > 0 &&
	    !KheLimitBusyTimesConstraintAllowZero(lbtc) &&
	    KheConstraintWeight((KHE_CONSTRAINT) lbtc) > 0 )
	  HaArrayAddLast(lbt_monitors, lbtm);
      }
    }
  }
  HaArraySort(lbt_monitors, &KheLimitBusyTimesMonitorDecreasingWeightCmp);

  /* handle each monitor */
  soln_changed = false;
  if( DEBUG1 )
    fprintf(stderr, "  %d monitors:\n", HaArrayCount(lbt_monitors));
  HaArrayForEach(lbt_monitors, lbtm, i)
  {
    if( DEBUG3 )
      KheMonitorDebug((KHE_MONITOR) lbtm, 2, 2, stderr);
    lbtc = KheLimitBusyTimesMonitorConstraint(lbtm);
    r = KheLimitBusyTimesMonitorResource(lbtm);
    rtm = KheResourceTimetableMonitor(soln, r);
    offset = KheLimitBusyTimesMonitorOffset(lbtm);
    min_limit = KheLimitBusyTimesConstraintMinimum(lbtc);
    for( j = 0;  j < KheLimitBusyTimesConstraintTimeGroupCount(lbtc);  j++ )
    {
      tg = KheLimitBusyTimesConstraintTimeGroup(lbtc, j, offset);
      if( KheTimeGroupTimeCount(tg) <= min_limit )
      {
	/* every time of tg is requested to be busy */
	for( k = 0;  k < KheTimeGroupTimeCount(tg);  k++ )
	{
	  t = KheTimeGroupTime(tg, k);
	  MakeBestAsst(soln, r, rtm, t, etm, &soln_changed);
	}
      }
    }
  }

  /* and return */
  KheSolnArenaEnd(soln, a);
  /* HaAre naDelete(a); */
  if( DEBUG1 )
    fprintf(stderr, "] KheSolnAssignRequestedResources returning %s\n",
      soln_changed ? "true" : "false");
  return soln_changed;
}
